"""
Cell quality module
"""
import ctypes
from pyansys import _cellqual

def CellQuality(grid):
    """
    DESCRIPTION
    Given a vtk unstructured grid, returns the minimum scaled jacobian cell 
    quality for each cell.  Unstructured grid can only contain the following
    elements
        tetrahedral
        pyramid
        wedge
        hexahedral
    
    Elements may be linear or quadradic
    
    """
    # Get cells and points from grid
    cells = grid.GetNumpyCells(ctypes.c_long)
    points = grid.GetNumpyPoints(ctypes.c_double)
    
    return _cellqual.CompScJac_quad(cells, points)




import os
import numpy as np
import warnings
import logging
import ctypes

from pyansys import _parsefull
from pyansys import _rstHelper

from pyansys import CDBparser


# attempt to load vtk
try:
    import vtkInterface
    import vtk
    vtkloaded = True

except:
    warnings.warn('Cannot load vtk\nWill be unable to display results.')
    vtkloaded = False


#==============================================================================
# Pointer information from ansys interface manual
#==============================================================================
#Individual element index table
e_table = ['ptrEMS', 'ptrENF', 'ptrENS', 'ptrENG', 'ptrEGR', 'ptrEEL', 
           'ptrEPL', 'ptrECR', 'ptrETH', 'ptrEUL', 'ptrEFX', 'ptrELF',
           'ptrEMN', 'ptrECD', 'ptrENL', 'ptrEHC', 'ptrEPT', 'ptrESF',
           'ptrEDI', 'ptrETB', 'ptrECT', 'ptrEXY', 'ptrEBA', 'ptrESV', 
           'ptrMNL']

"""
ptrEMS - pointer to misc. data
ptrENF - pointer to nodal forces
ptrENS - pointer to nodal stresses
ptrENG - pointer to volume and energies
ptrEGR - pointer to nodal gradients
ptrEEL - pointer to elastic strains
ptrEPL - pointer to plastic strains
ptrECR - pointer to creep strains
ptrETH - pointer to thermal strains
ptrEUL - pointer to euler angles
ptrEFX - pointer to nodal fluxes
ptrELF - pointer to local forces
ptrEMN - pointer to misc. non-sum values
ptrECD - pointer to element current densities
ptrENL - pointer to nodal nonlinear data
ptrEHC - pointer to calculated heat generations
ptrEPT - pointer to element temperatures
ptrESF - pointer to element surface stresses
ptrEDI - pointer to diffusion strains
ptrETB - pointer to ETABLE items(post1 only
ptrECT - pointer to contact data
ptrEXY - pointer to integration point locations
ptrEBA - pointer to back stresses
ptrESV - pointer to state variables
ptrMNL - pointer to material nonlinear record


"""


class FullReader(object):
    """
    Object to store the results of an ANSYS full file

    NOTES:
    Currently only supports symmetric and real stiffness matrices as well as
    non-lumped mass matrices.
    
    """
    
    def __init__(self, filename):
        """
        Loads full header on initialization
        
        See ANSYS programmer's reference manual full header section for
        definitions of each header.
        
        """
        
        # check if file exists
        if not os.path.isfile(filename):
            raise Exception('{:s} not found'.format(filename))
        
        self.filename = filename
        self.header = _parsefull.ReturnHeader(filename)
        
        #// Check if lumped (item 11)
        if self.header[11]:
            raise Exception("Unable to read a lumped mass matrix.  Terminating.")
    
        # Check if arrays are unsymmetric (item 14)
        if self.header[14]:
            raise Exception ("Unable to read an unsymmetric mass/stiffness matrix.")


# Dead setting
#        sort : bool, optional
#            By default, this setting sorts the rows and columns such that the 
#            nodes are in order.  ANSYS stores the mass and stiffness matrices 
#            such that the bandwidth of the arrays is minimized.  Therefore, to
#            minimize the bandwidth of the arrays, make this setting False.
    def LoadKM(self, as_sparse=True, utri=True):
        """
        Load and construct mass and stiffness matrices from an ANSYS full file.

        Parameters
        ----------
        as_sparse : bool, optional
            Outputs the mass and stiffness matrices as scipy csc sparse arrays
            when True by default.
            
        utri : bool, optional
            Outputs only the upper triangle of both the mass and stiffness
            arrays.
        
        Returns
        -------
        dof_ref : (n x 2) np.int32 array
            This array contains the node and degree corresponding to each row 
            and column in the mass and stiffness matrices.  When the sort 
            parameter is set to True this array will be sorted by node number
            first and then by the degree of freedom.  In a 3 DOF analysis the 
            intergers will correspond to:
            0 - x
            1 - y
            2 - z
            
        k : (n x n) np.float or scipy.csc array
            Stiffness array
            
        m : (n x n) np.float or scipy.csc array
            Mass array
        
        Notes
        -----
        
        
        """
        # check file still exists
        if not os.path.isfile(self.filename):
            raise Exception('{:s} not found'.format(self.filename))
            
        # see if 
        if as_sparse:
            try:
                from scipy.sparse import csc_matrix, coo_matrix
            except:
                raise Exception('Unable to load scipy, matricies will be full')
                as_sparse = False
        
        # Get header details
        neqn = self.header[2];    #  Number of equations
        ntermK = self.header[9];  # number of terms in stiffness matrix
        ptrSTF = self.header[19]; # Location of stiffness matrix
        ptrMAS = self.header[27]; # Location in file to mass matrix
        nNodes = self.header[33]; # Number of nodes considered by assembly
        ntermM = self.header[34]; # number of terms in mass matrix
        ptrDOF = self.header[36]; # pointer to DOF info
        
        # get details for reading the mass and stiffness arrays
        node_info = _rstHelper.FullNodeInfo(self.filename, ptrDOF, nNodes, 
                                            neqn)
        
        nref, dref, index_arr, const, ndof = node_info
        dof_ref = np.vstack((nref, dref)).T # stack these references
        
        # Read k and m blocks (see help(ReadArray) for block description)
        k_block = _rstHelper.ReadArray(self.filename, ptrSTF, ntermK, neqn, 
                                       index_arr)
        
        m_block = _rstHelper.ReadArray(self.filename, ptrMAS, ntermM, neqn, 
                                       index_arr)        
        k_diag = k_block[3]
        k_data_diag = k_block[4]
    
        m_diag = m_block[3]
        m_data_diag = m_block[4]
        
        # assemble data
        if utri:
            # stiffness matrix
            krow = np.hstack((k_block[1], k_diag)) # row and diag
            kcol = np.hstack((k_block[0], k_diag)) # col and diag
            kdata= np.hstack((k_block[2], k_data_diag)) # data and diag

            # mass matrix
            mrow = np.hstack((m_block[1], m_diag)) # row and diag
            mcol = np.hstack((m_block[0], m_diag)) # col and diag
            mdata= np.hstack((m_block[2], m_data_diag)) # data and diag
        
        else:
            # stiffness matrix
            krow = np.hstack((k_block[0], k_block[1], k_diag)) # row, col and diag
            kcol = np.hstack((k_block[1], k_block[0], k_diag)) # col and diag
            kdata= np.hstack((k_block[2], k_block[2], k_data_diag)) # data and diag

            # mass matrix
            mrow = np.hstack((m_block[0], m_block[1], m_diag)) # row, col and diag
            mcol = np.hstack((m_block[1], m_block[0], m_diag)) # col and diag
            mdata= np.hstack((m_block[2], m_block[2], m_data_diag)) # data and diag

                
        # number of dimentions
        ndim = nref.size

        # output as a sparse matrix
        if as_sparse:
            
            k = coo_matrix((ndim,)*2)
            k.data = kdata # data has to be set first
            k.row = krow
            k.col = kcol
            
            # convert to csc matrix (generally faster for sparse solvers)
            k = csc_matrix(k)

            m = coo_matrix((ndim,)*2)
            m.data = mdata
            m.row = mrow
            m.col = mcol

            # convert to csc matrix (generally faster for sparse solvers)
            m = csc_matrix(m)
            
        else:
            k = np.zeros((ndim,)*2)
            k[krow, kcol] = kdata

            m = np.zeros((ndim,)*2)
            m[mrow, mcol] = mdata
            
        # store if constrained and number of degrees of freedom per node
        self.const = const < 0
        self.ndof  = ndof
        

        return dof_ref, k, m
        

class ResultReader(object):
    """
    Object to control the reading of ANSYS results written to fortran file
    """
    
    def __init__(self, filename, logger=False, load_geometry=True):
        """
        Loads basic result information from result file and initializes result
        object.
    
        Parameters
        ----------
        filename : string
            Filename of the ANSYS binary result file.
            
        logger : bool, optional
            Enables logging if True.  Debug feature.
            
        load_geometry : bool, optional
            Loads geometry using vtk by default
            
        Returns
        -------
        None
        
        """

        # set logging level depending on settings
        if logger:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.CRITICAL)

        # store logger pointer
        self.logger = logging.getLogger(__name__)

        # Store filename result header items
        self.filename = filename
        self.resultheader = GetResultInfo(filename)
        
        # Get the total number of results and log it
        self.nsets = len(self.resultheader['rpointers'])
        string = 'There are {:d} results in this file'.format(self.nsets)
        self.logger.debug(string)
        
        # Get indices to resort nodal and element results
        self.sidx = np.argsort(self.resultheader['neqv'])
        self.sidx_elem = np.argsort(self.resultheader['eeqv'])

        # Store node numbering in ANSYS
        self.nnum = self.resultheader['neqv'][self.sidx]
        self.enum = self.resultheader['eeqv'][self.sidx_elem]
        
        # Store time values for later retrival
        self.GetTimeValues()
        
        # store geometry for later retrival
        if load_geometry:
            self.StoreGeometry()
           
        if self.resultheader['nSector'] and load_geometry:
            self.iscyclic = True
            
            # Add cyclic properties
            self.AddCyclicProperties()
            
            
    def AddCyclicProperties(self):
        """ Adds cyclic properties to result object """
        # ansys's duplicate sector contains nodes from the second half of the node
        # numbering
        # 
        # ansys node numbering for the duplicate sector is
        # nnum.min() + nnum.max()
        # where nnum is the node numbering for the master sector
        
#        if not vtkloaded:
#            raise Exception('Unable to add ')
        
        # master sector max node number
        num_master_max = self.nnum[int(self.nnum.size/2) - 1]
        
        # identify master and duplicate cyclic sectors
        cells = np.arange(self.enum.size)
        dup_cells = np.where(np.any(self.geometry['elem'] > num_master_max, 1))[0]
        mas_cells = np.setdiff1d(cells, dup_cells)
        self.sector = self.uGrid.ExtractSelectionCells(mas_cells)
        dup_sector = self.uGrid.ExtractSelectionCells(dup_cells)
        
        # Store the indices of the master and duplicate nodes
        self.mas_ind = self.sector.GetPointScalars('vtkOriginalPointIds')
        self.dup_ind = dup_sector.GetPointScalars('vtkOriginalPointIds')
        
        # store cyclic node numbers
        self.cyc_nnum = self.nnum[self.mas_ind]

        # create full rotor
        nSector = self.resultheader['nSector']
        
        # Copy and translate mesh
        vtkappend = vtk.vtkAppendFilter()
        rang = 360.0/nSector
        for i in range(nSector):
                
            # Transform mesh
            sector = self.sector.Copy()
            sector.RotateZ(rang*i)
        
            vtkappend.AddInputData(sector)
                
        # Combine meshes and add VTK_Utilities functions
        #vtkappend.MergePointsOn()
        vtkappend.Update()
        self.rotor = vtkappend.GetOutput()
        vtkInterface.GridAddExtraFunctions(self.rotor)
        
        
    def GetCyclicNodalResult(self, rnum):
        """
        Returns the nodal result given a cumulative result index.
        
        Parameters
        ----------
        rnum : interger
            Cumulative result number.  Zero based indexing.
            
        
        Returns
        -------
        result : numpy.complex128 array
            Result is (nnod x numdof), nnod is the number of nodes in a sector
            and numdof is the number of degrees of freedom.
            
        Notes
        -----
        Node numbers correspond to self.cyc_nnum, where self is this result 
        object
            
        """
        if not self.iscyclic:
            raise Exception('Result file does not contain cyclic results.')

        # get the nodal result
        r = self.GetNodalResult(rnum)
        
        return r[self.mas_ind] + r[self.dup_ind]*1j
        
    
    def PlotCyclicNodalResult(self, rnum, phase=0, comp='norm', as_abs=False, label='',
                              expand=True):
        """
        Plots a nodal result from a cyclic analysis.
        
        Parameters
        ----------
        rnum : interger
            Cumulative result number.  Zero based indexing.
            
        phase : float, optional
            Shifts the phase of the solution.
            
        comp : string, optional
            Display component to display.  Options are 'x', 'y', 'z', and
            'norm', corresponding to the x directin, y direction, z direction,
            and the combined direction (x**2 + y**2 + z**2)**0.5
            
        as_abs : bool, optional
            Displays the absolute value of the result.
            
        label: string, optional
            Annotation string to add to scalar bar in plot.
            
        expand : bool, optional
            Expands the solution to a full rotor when True.  Enabled by 
            default.  When disabled, plots the maximum response of a single
            sector of the cyclic solution in the component of interest.
        
        Returns
        -------
        cpos : list
            Camera position from vtk render window.
            
        Notes
        -----
        None
        
        """
        
        if 'hindex' not in self.resultheader:
            raise Exception('Result file does not contain cyclic results')
        
        # harmonic index and number of sectors
        hindex = self.resultheader['hindex'][rnum]
        nSector = self.resultheader['nSector']

        # get the nodal result
        r = self.GetCyclicNodalResult(rnum)
        
        alpha = (2*np.pi/nSector)

        if expand:
            grid = self.rotor
            d = np.empty((self.rotor.GetNumberOfPoints(), 3))
            n = self.sector.GetNumberOfPoints()
            for i in range(nSector):
                # adjust the phase of the result
                sec_sol = np.real(r)*np.cos(i*hindex*alpha + phase) -\
                          np.imag(r)*np.sin(i*hindex*alpha + phase)
                          
                # adjust the "direction" of the x and y vectors as they're being
                # rotated
                s_x = sec_sol[:, 0]*np.cos(alpha*i + phase) -\
                      sec_sol[:, 1]*np.sin(alpha*i + phase)
                s_y = sec_sol[:, 0]*np.sin(alpha*i + phase) +\
                      sec_sol[:, 1]*np.cos(alpha*i + phase)
                sec_sol[:, 0] = s_x
                sec_sol[:, 1] = s_y
                
                d[i*n:(i + 1)*n] = sec_sol
        
        else:
            # plot the max response for a single sector
            grid = self.sector

            n = np.sum(r*r, 1)
                
            # rotate the response based on the angle to maximize the highest
            # responding node
            angle = np.angle(n[np.argmax(np.abs(n))])
            d = np.real(r)*np.cos(angle + phase) -\
                np.imag(r)*np.sin(angle + phase)   
                
            d = -d
                
        # Process result
        if comp == 'x':
            d = d[:, 0]
            stitle = 'X {:s}'.format(label)
            
        elif comp == 'y':
            d = d[:, 1]
            stitle = 'Y {:s}'.format(label)
            
        elif comp == 'z':
            d = d[:, 2]
            stitle = 'Z {:s}'.format(label)
            
        else:
            # Normalize displacement
            d = d[:, :3]
            d = (d*d).sum(1)**0.5
            
            stitle = 'Normalized\n{:s}'.format(label)
            
        if as_abs:
            d = np.abs(d)
        
        # setup text
        ls_table = self.resultheader['ls_table']
        freq = self.GetTimeValues()[rnum]
        text  = 'Cumulative Index: {:3d}\n'.format(ls_table[rnum, 2])
        text += 'Loadstep:         {:3d}\n'.format(ls_table[rnum, 0])
        text += 'Substep:          {:3d}\n'.format(ls_table[rnum, 1])
        text += 'Harmonic Index:   {:3d}\n'.format(hindex)
        text += 'Frequency:      {:10.4f} Hz'.format(freq)

        # plot        
        plobj = vtkInterface.PlotClass()
        plobj.AddMesh(grid, scalars=d, stitle=stitle, flipscalars=True,
                      interpolatebeforemap=True)
        plobj.AddText(text, fontsize=20)
        plobj.Plot(); del plobj


    def ResultsProperties(self):
        """
        Logs results available in the result file and returns a dictionary
        of available results
        
        Logging must be enabled for the results of the check to be shown in the
        console.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        result_check : dictionary
            Dictionary indicating the availability of results.
            
        Notes
        -----
        None
        
        """
        
        # check number of results
        self.logger.debug('There are {:d} results in this file'.format(self.nsets))

        if self.resultheader['nSector'] > 1:
            self.logger.debug('Contains results from a cyclic analysis with:')
            self.logger.debug('\t{:d} sectors'.format(self.resultheader['nSector']))

        return {'Number of Results': self.nsets}


    def PlotNodalResult(self, rnum, comp='norm', as_abs=False, label=''):
        """
        Plots a nodal result.  
        
        Parameters
        ----------
        rnum : interger
            Cumulative result number.  Zero based indexing.
            
        comp : string, optional
            Display component to display.  Options are 'x', 'y', 'z', and
            'norm', corresponding to the x directin, y direction, z direction,
            and the combined direction (x**2 + y**2 + z**2)**0.5
            
        as_abs : bool, optional
            Displays the absolute value of the result.
            
        label: string, optional
            Annotation string to add to scalar bar in plot.
        
        Returns
        -------
        cpos : list
            Camera position from vtk render window.
            
        Notes
        -----
        None
        
        """
        if not vtkloaded:
            raise Exception('Cannot plot without VTK')

        # Load result from file
        result = self.GetNodalResult(rnum)

        # Process result
        if comp == 'x':
            d = result[:, 0]
            stitle = 'X {:s}'.format(label)
            
        elif comp == 'y':
            d = result[:, 1]
            stitle = 'Y {:s}'.format(label)
            
        elif comp == 'z':
            d = result[:, 2]
            stitle = 'Z {:s}'.format(label)
            
        else:
            # Normalize displacement
            d = result[:, :3]
            d = (d*d).sum(1)**0.5
            
            stitle = 'Normalized\n{:s}'.format(label)
            
        if as_abs:
            d = np.abs(d)
        
        # Generate plot
        text = 'Result {:d} at {:f}'.format(rnum + 1, self.tvalues[rnum])
        plobj = vtkInterface.PlotClass()
        plobj.AddMesh(self.uGrid, no_copy=True, scalars=d, stitle=stitle, 
                      flipscalars=True, interpolatebeforemap=True)
        plobj.AddText(text)
        cpos = plobj.Plot();  # store camera position
        del plobj

        return cpos
    
    
    def GetTimeValues(self):
        """
        SUMMARY
        Returns table of time values for results.  For a modal analysis, this
        corresponds to the frequencies of each mode.
        
        
        INPUTS
        None
            
            
        OUTPUTS
        tvalues (np.float64 array)
        
        """
        
        endian = self.resultheader['endian']
        ptrTIMl = self.resultheader['ptrTIMl']
        
        # Load values if not already stored
        if not hasattr(self, 'tvalues'):
    
            # Seek to start of time result table
            f = open(self.filename, 'rb')
    
            f.seek(ptrTIMl*4 + 8)
            self.tvalues = np.fromfile(f, endian + 'd', self.nsets)
            
            f.close()
        
        return self.tvalues


    def GetNodalResult(self, rnum, sort=True):
        """
        Returns the nodal result for a result number
        
        Parameters
        ----------
        rnum : interger
            Cumulative result number.  Zero based indexing.
            
        sort : bool, optional
            Resorts the results so that the results correspond to the sorted
            node numbering (self.nnum) (default).  If left unsorted, results 
            correspond to the nodal equivalence array self.resultheader['neqv']
        
        Returns
        -------
        result : numpy.float array
            Result is (nnod x numdof), or number of nodes by degrees of freedom
            
        Notes
        -----
        None
            
        """
        # Get info from result header
        endian = self.resultheader['endian']
        numdof = self.resultheader['numdof']
        nnod = self.resultheader['nnod']
        rpointers = self.resultheader['rpointers']
        
        # Check if result is available
        if rnum > self.nsets - 1:
            raise Exception('There are only {:d} results in the result file.'.format(self.nsets))
        
        # Read a result
        f = open(self.filename, 'rb')
        
        # Seek to result table and to get pointer to DOF results of result table
        f.seek((rpointers[rnum] + 12)*4) # item 11
        ptrNSLl = np.fromfile(f, endian + 'i', 1)[0]
        
        # Seek and read DOF results
        f.seek((rpointers[rnum] + ptrNSLl + 2)*4)
        nitems = nnod*numdof
        result = np.fromfile(f, endian + 'd', nitems)
        
        f.close()
        
        # Reshape to number of degrees of freedom
        r = result.reshape((-1, numdof))
        
        # if using a cyclic coordinate system
        if self.resultheader['csCord']:
            # compute sin and cos angles
            if not hasattr(self, 's_angle'):
                # angle of each point
                angle = np.arctan2(self.geometry['nodes'][:, 1], self.geometry['nodes'][:, 0])
                angle = angle[np.argsort(self.sidx)]
                self.s_angle = np.sin(angle)#[self.sidx]
                self.c_angle = np.cos(angle)#[self.sidx]
                
            rx = r[:, 0]*self.c_angle - r[:, 1]*self.s_angle
            ry = r[:, 0]*self.s_angle + r[:, 1]*self.c_angle
            r[:, 0] = rx
            r[:, 1] = ry
        
        # Reorder based on sorted indexing and return
        if sort:
            r = r.take(self.sidx, 0)
            
        return r
    
    
    def StoreGeometry(self):
        """ Stores the geometry from the result file """
        
        f = open(self.filename, 'rb')
        f.seek((self.resultheader['ptrGEO'] + 2)*4)
        geotable = np.fromfile(f, self.resultheader['endian'] + 'i', 80)
        geotable.tolist()
        
        ptrLOC = geotable[26]
        
        #==============================================================================
        # Node information
        #==============================================================================
        nnod = self.resultheader['nnod']
        nnum = np.empty(nnod, np.int32)
        nloc = np.empty((nnod, 6), np.float)
        _rstHelper.LoadNodes(self.filename, ptrLOC, nnod, nloc, nnum)
        
        #======================================================================
        # Element Information
        #======================================================================
        nelm = geotable[4]
        ptrEID = geotable[28]
        maxety = geotable[1]
        
        # pointer to the element type index table
        ptrETYP = geotable[20]
        f.seek((ptrETYP + 2)*4)
        e_type_table = np.fromfile(f, self.resultheader['endian'] + 'i', maxety)
        
        # store information for each element type
        # make these arrays large so you can reference a value via element type numbering
#        etype_arr = np.empty(10000, np.int32)
        nodelm = np.empty(10000, np.int32) # number of nodes for this element type
        nodfor = np.empty(10000, np.int32) # number of nodes per element having nodal forces
        nodstr = np.empty(10000, np.int32) # number of nodes per element having nodal stresses
        etype_ID = np.empty(maxety, np.int32)
        ekey = []
        for i in range(maxety):
            f.seek((ptrETYP + e_type_table[i] + 2)*4)
            einfo = np.fromfile(f, self.resultheader['endian'] + 'i', 2)
            etype_ref = einfo[0]
            etype_ID[i] = einfo[1]
            ekey.append(einfo)
            
            f.seek((ptrETYP + e_type_table[i] + 2 + 60)*4)
            nodelm[etype_ref] = np.fromfile(f, self.resultheader['endian'] + 'i', 1)
        
            f.seek((ptrETYP + e_type_table[i] + 2 + 62)*4)
            nodfor[etype_ref] = np.fromfile(f, self.resultheader['endian'] + 'i', 1)
        
            f.seek((ptrETYP + e_type_table[i] + 2 + 93)*4)
            nodstr[etype_ref] = np.fromfile(f, self.resultheader['endian'] + 'i', 1)
        
        # store element table data
        self.element_table = {'nodelm': nodelm,
                              'nodfor': nodfor,
                              'nodstr': nodstr}
        
        # get the element description table
        f.seek((ptrEID + 2)*4)
        e_disp_table = np.empty(nelm, np.int32)
        e_disp_table[:] = np.fromfile(f, self.resultheader['endian'] + 'i8', nelm)
        
        # get pointer to start of element table and adjust element pointers
        ptr = ptrEID + e_disp_table[0]
        e_disp_table -= e_disp_table[0]


        f.close()
        
        # The following is stored for each element
        """
        mat     - material reference number
        type    - element type number
        real    - real constant reference number
        secnum  - section number
        esys    - element coordinate system
        death   - death flat (1 live, 0 dead)
        solidm  - solid model reference
        shape   - coded shape key
        elnum   - element number
        baseeid - base element number
        NODES   - node numbers defining the element
        """
        
        
        # allocate memory for this (a maximum of 21 points per element)
        etype = np.empty(nelm, np.int32)

        elem = np.empty((nelm, 20), np.int32)
        elem[:] = -1
        
        # load elements
        _rstHelper.LoadElements(self.filename, ptr, nelm, e_disp_table, elem, 
                                etype)
        enum = self.resultheader['eeqv']
        
        # store geometry dictionary
        self.geometry = {'nnum': nnum,
                         'nodes': nloc,
                         'etype': etype,
                         'elem': elem,
                         'enum': enum,
                         'ekey': np.asarray(ekey, ctypes.c_long)}
        
        
        # store the reference array
        cells, offset, cell_type, self.numref = CDBparser.Parse(self.geometry, True)

        # Create vtk object if vtk installed
        if vtkloaded:
            nodes = nloc[:, :3]
            self.uGrid = vtkInterface.MakeuGrid(offset, cells, cell_type, nodes)
        
        # get edge nodes            
        nedge = nodstr[etype].sum()
        self.edge_idx = np.empty(nedge, np.int32)
        _rstHelper.AssembleEdges(nelm, etype, elem, self.numref.astype(np.int32), 
                                 self.edge_idx, nodstr)
            
        
        # store edge node numbers and indices to the node array
        self.edge_node_num_idx = np.unique(self.edge_idx)

        # catch the disassociated node bug
#        try:
#            self.edge_node_num = self.geometry['nnum'][self.edge_node_num_idx]
#        except:
#            logging.warning('unable to generate edge_node_num')
            
    
    def NodalStress(self, rnum):
        """
        Retrives the component stresses for each node in the solution.
        
        The order of the results corresponds to the sorted node numbering.
        
        This algorthim, like ANSYS, computes the nodal stress by averaging the
        stress for each element at each node.  Due to the discontinunities
        across elements, stresses will vary based on the element they are
        evaluated from.
        
        Parameters
        ----------
        rnum : interger
            Result set to load using zero based indexing.
        
        Returns
        -------
        stress : array
            Stresses at Sx Sy Sz Sxy Syz Sxz averaged at each corner node.  
            For the corresponding node numbers, see "edge_node_num"
            
        Notes
        -----
        None
            
        """
        #%% debug cell
        # Get the header information from the header dictionary
        endian = self.resultheader['endian']
        rpointers = self.resultheader['rpointers']
        nelm = self.resultheader['nelm']
        nodstr = self.element_table['nodstr']
        etype = self.geometry['etype']
        
        # Check if result is available
        if rnum > self.nsets - 1:
            raise Exception('There are only {:d} results in the result file.'.format(self.nsets))
        
        # Read a result
        f = open(self.filename, 'rb')
        
        # Seek to result table and to get pointer to DOF results of result table
        f.seek((rpointers[rnum] + 13)*4) # item 12
        
        # 32-bit pointer to element solution
        ptrESL = np.fromfile(f, endian + 'i', 1)[0]
        
        if not ptrESL:
            f.close()
            raise Exception('No element solution in result set {:d}'.format(rnum + 1))
        
        # Seek to element result header
        element_rst_ptr = rpointers[rnum] + ptrESL + 2
        f.seek(element_rst_ptr*4)
        
        # element index table
        ele_ind_table = np.fromfile(f, endian + 'i8', nelm).astype(np.int32)
        ele_ind_table += element_rst_ptr
        
        # Each element header contains 25 records for the individual results
        table_index = e_table.index('ptrENS')
        
        # check number of records to read (differs with each version)
        f.seek((ele_ind_table[0] + table_index)*4)
        ptrENS = np.fromfile(f, endian + 'i', 1)[0]

        nnode_elem = nodstr[etype[0]]
        f.seek((ele_ind_table[0] + ptrENS - 2)*4)
        nitem = np.fromfile(f, endian + 'i', 1)[0]/nnode_elem


        nstresses = self.edge_idx.size
        stresses = np.empty((nstresses, 6), np.float32)
        
        #%% debug cell 2
        c = 0
        for i in range(len(ele_ind_table)):
            # Element nodal stresses, ptrENS, is the third item in the table
            f.seek((ele_ind_table[i] + table_index)*4)
            ptrENS = np.fromfile(f, endian + 'i', 1)[0]
        
            # read the stresses evaluated at the intergration points or nodes
            nnode_elem = nodstr[etype[i]]
            
            f.seek((ele_ind_table[i] + ptrENS)*4)
            stress = np.fromfile(f, endian + 'f', nnode_elem*nitem).reshape((-1, nitem))#[:, sidx]

            # store stresses
            stresses[c:c + nnode_elem] = stress[:, :6]
            c += nnode_elem
            
        # close file
        f.close()
        
        
        # Average the stresses for each element at each node
#        enode = self.edge_node_num_idx
#        s_node = np.empty((enode.size, 6), np.float32)
#        for i in range(6):
#            s_node[:, i] = np.bincount(self.edge_idx, weights=stresses[:, i])[enode]
#        ntimes = np.bincount(self.edge_idx)[enode]
#        s_node /= ntimes.reshape((-1, 1))



        # grabe element results from binary
        enode = self.edge_node_num_idx
        ntimes = np.bincount(self.edge_idx)[enode]
        
        nnod = self.resultheader['nnod']
        ele_data_arr = np.zeros((nnod, 6), np.float32)
        _rstHelper.LoadStress(self.filename, table_index, 
                              ele_ind_table, nodstr, etype, 
                              nitem, ele_data_arr, self.edge_idx)
        
        s_node = ele_data_arr[enode]
        s_node /= ntimes.reshape((-1, 1))
        
        return s_node
        
    
    def PlotNodalStress(self, rnum, stype):
        """
        Plots the stresses at each node in the solution.
        
        The order of the results corresponds to the sorted node numbering.
        This algorthim, like ANSYS, computes the node stress by averaging the
        stress for each element at each node.  Due to the discontinunities
        across elements, stresses will vary based on the element they are
        evaluated from.
        
        Parameters
        ----------
        rnum : interger
            Result set using zero based indexing.
        stype : string
            Stress type from the following list: [Sx Sy Sz Sxy Syz Sxz]
        
        Returns
        -------
        None
            
        Notes
        -----
        None
            
        """
        
        stress_types = ['Sx', 'Sy', 'Sz', 'Sxy', 'Syz', 'Sxz', 'Seqv']
        if stype not in stress_types:
            raise Exception("Stress type not in \n ['Sx', 'Sy', 'Sz', 'Sxy', 'Syz', 'Sxz']")

        sidx = ['Sx', 'Sy', 'Sz', 'Sxy', 'Syz', 'Sxz'].index(stype)

        # create a zero array for all nodes
        stress = np.zeros(self.resultheader['nnod'])
        
        # Populate with nodal stress at edge nodes
        stress[self.edge_node_num_idx] = self.NodalStress(rnum)[:, sidx]
        stitle = 'Nodal Stress\n{:s}'.format(stype)
        
        # Generate plot
        plobj = vtkInterface.PlotClass()
        plobj.AddMesh(self.uGrid, scalars=stress, stitle=stitle, no_copy=True,
                      flipscalars=True, interpolatebeforemap=True)
        
        text = 'Result {:d} at {:f}'.format(rnum + 1, self.tvalues[rnum])
        plobj.AddText(text)
        
        cpos = plobj.Plot() # store camera position
        del plobj
        
        return cpos


def GetResultInfo(filename):
    """
    Returns pointers used to access results from an ANSYS result file.

    Parameters
    ----------
    filename : string
        Filename of result file.

    Returns
    -------
    resultheader : dictionary
    
    """
    
    f = open(filename, 'rb')

    # initialize result header dictionary
    resultheader = {}

    # Check if big or small endian
    endian = '<'
    if np.fromfile(f, dtype='<i', count=1) != 100:
        
        # Check if big enos
        f.seek(0)
        if np.fromfile(f, dtype='>i', count=1) == 100:
            endian = '>'

        # Otherwise, it's probably not a result file
        else:
            raise Exception('Unable to determine endian type.\n\n' +\
                            'File is possibly not a result file.')

    resultheader['endian'] = endian

    # Read standard header
    f.seek(0);
    
    # Get ansys version
    f.seek(11*4)
    version = f.read(4)[::-1]

    try:
        resultheader['verstring'] = version
        resultheader['mainver'] = int(version[:2])
        resultheader['subver'] = int(version[-1])
    except:
        warnings.warn('Unable to parse version')
        resultheader['verstring'] = 'unk'
        resultheader['mainver'] = 15
        resultheader['subver'] = 0
        
    
    #======================
    # Read .RST FILE HEADER 
    #======================
    # 100 is size of standard header, plus extras, 3 is location of pointer in table
    f.seek(105*4)
    rheader = np.fromfile(f, endian + 'i', count=55)
    
    # Number of nodes (item 3)
    resultheader['nnod'] = rheader[2]

    # Number of elements (item 6)
    resultheader['nelm'] = rheader[6]
    
    # Number of degrees of freedom (item 5)
    resultheader['numdof'] = rheader[4]
    
    # Number of sets of results
    resultheader['nsets'] = rheader[8]
    
    # Pointer to results table (item 11)
    ptrDSIl = rheader[10]
    
    # Pointer to the table of time values for a load step (item 12)
    resultheader['ptrTIMl'] = rheader[11]
    
    # pointer to load step table (item 13)
    ptrLSP = rheader[12]    
    
    # pointer to element equivalence table (item 14)
    ptrELM = rheader[13]
    
    # pointer to nodal equivalence table (item 15)
    ptrNODl = rheader[14]
    
    # pointer to geometry information (item 16)
    resultheader['ptrGEO'] = rheader[15]
    
    # pointer to cyclic symmetry nodal-diameters for each load step
    resultheader['ptrCYC'] = rheader[16]
    
    # number of sectors (item 21)
    resultheader['nSector'] = rheader[20]
    
    # cyclic symmetry coordinate system
    resultheader['csCord'] = rheader[21]
    
    # Read nodal equivalence table
    f.seek((ptrNODl + 2)*4) # Start of pointer, then empty, then data
    resultheader['neqv'] = np.fromfile(f, endian + 'i', count=resultheader['nnod'])
    
    # Read nodal equivalence table
    f.seek((ptrELM + 2)*4) # Start of pointer, then empty, then data
    resultheader['eeqv'] = np.fromfile(f, endian + 'i', count=resultheader['nelm'])
    
    # Read table of pointers to locations of results
    f.seek((ptrDSIl + 2)*4) # Start of pointer, then empty, then data
    rpointers = np.fromfile(f, endian + 'i', count=resultheader['nsets'])
    resultheader['rpointers'] = rpointers
    
    # load harmonic index of each result
    if resultheader['ptrCYC']:
        f.seek((resultheader['ptrCYC'] + 2)*4)
        resultheader['hindex'] = np.fromfile(f, endian + 'i',
                    count=resultheader['nsets'])

    # load step table with columns:
    # [loadstep, substep, and cumulative]        
    f.seek((ptrLSP + 2)*4) # Start of pointer, then empty, then data
    table = np.fromfile(f, endian + 'i', count=resultheader['nsets']*3)
    resultheader['ls_table'] = table.reshape((-1, 3))
    
    f.close()
    
    return resultheader


def Unique_Rows(a):
    """ Returns unique rows of a and indices of those rows """
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx, ridx = np.unique(b, return_index=True, return_inverse=True)
    
    return idx, ridx


def delete_row_csc(mat, i):
    """ remove a row from a csc matrix """
#    if not isinstance(mat, scipy.sparse.csr_matrix):
#        raise ValueError("works only for CSR format -- use .tocsr() first")
    n = mat.indptr[i+1] - mat.indptr[i]
    if n > 0:
        mat.data[mat.indptr[i]:-n] = mat.data[mat.indptr[i+1]:]
        mat.data = mat.data[:-n]
        mat.indices[mat.indptr[i]:-n] = mat.indices[mat.indptr[i+1]:]
        mat.indices = mat.indices[:-n]
    mat.indptr[i:-1] = mat.indptr[i+1:]
    mat.indptr[i:] -= n
    mat.indptr = mat.indptr[:-1]
    mat._shape = (mat._shape[0]-1, mat._shape[1])


# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

import numpy as np
cimport numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

import ctypes               
         
# Type defintion for an unsigned 8-bit
ctypedef unsigned char uint8
          
# VTK numbering for vtk cells
cdef uint8 vtkhexnum = 12
cdef uint8 vtkpyrnum = 14
cdef uint8 vtktetnum = 10
cdef uint8 vtkwegnum = 13
                   
# Quadradic elements
cdef int vtkquadtetnum = 24 # VTK_QUADRATIC_TETRA = 24
cdef int vtkquadpyrnum = 27 # VTK_QUADRATIC_PYRAMID = 27
cdef int vtkquadwegnum = 26 # VTK_QUADRATIC_WEDGE = 26
cdef int vtkquadhexnum = 25 # VTK_QUADRATIC_HEXAHEDRON = 25                        
                   
# ANSYS element type definitions
cdef int [4] typeA

# Legacy mixed elements
typeA[0] = 45
typeA[1] = 95

# Current mixed elements
typeA[2] = 185
typeA[3] = 186

# Tetrahedrals (legacy and current)
cdef int [4] typeB
typeB[0] = 92
typeB[1] = 187
                   
#==============================================================================
# Store elements
#==============================================================================
cdef inline void StoreTet_TypeB(long [::1] offset, long *ecount, long *ccount, 
                          long [::1] cells, uint8 [::1] cell_type,
                          long [::1] numref, int [:, ::1] elem, int i, int lin):
    """
    Stores tetrahedral element in vtk arrays.  ANSYS elements are ordered
    the same as vtk elements.
    
    VTK DOCUMENTATION
    The tetrahedron is defined by the four points (0-3); where (0,1,2) is the
    base of the tetrahedron which, using the right hand rule, forms a triangle
    whose normal points in the direction of the fourth point.
    
    For quadradic
    The cell includes a mid-edge node on each of the size edges of the 
    tetrahedron. The ordering of the ten points defining the cell is point ids
    (0-3,4-9) where ids 0-3 are the four tetra vertices; and point ids 4-9 are
    the midedge nodes between
    (0,1), (1,2), (2,0), (0,3), (1,3), and (2,3)

    """
    # Populate offset array
    offset[ecount[0]] = ccount[0]
    
    if lin:
        cells[ccount[0]] = 4; ccount[0] += 1
        
        # Populate cell array while renumbering nodes
        for j in range(4):
            cells[ccount[0]] = numref[elem[i, j]]; ccount[0] += 1
        
        # Populate cell type array
        cell_type[ecount[0]] = vtktetnum
        
    else:
        cells[ccount[0]] = 10; ccount[0] += 1
        
        # Populate cell array while renumbering nodes
        for j in range(10):
            cells[ccount[0]] = numref[elem[i, j]]; ccount[0] += 1
        
        # Populate cell type array
        cell_type[ecount[0]] = vtkquadtetnum
        
    # increment element counter
    ecount[0] += 1
    

cdef inline void StoreTet(long [::1] offset, long *ecount, long *ccount, 
                          long [::1] cells, uint8 [::1] cell_type,
                          long [::1] numref, int [:, ::1] elem, int i, int lin):
    """
    Stores tetrahedral element in vtk arrays.  ANSYS elements are ordered
    the same as vtk elements.
    
    VTK DOCUMENTATION:
    Linear
    The tetrahedron is defined by the four points (0-3); where (0,1,2) is the
    base of the tetrahedron which, using the right hand rule, forms a triangle
    whose normal points in the direction of the fourth point.
 
    Quadradic   
    The cell includes a mid-edge node on each of the size edges of the 
    tetrahedron. The ordering of the ten points defining the cell is point ids
    (0-3,4-9) where ids 0-3 are the four tetra vertices; and point ids 4-9 are
    the midedge nodes between
    (0,1), (1,2), (2,0), (0,3), (1,3), and (2,3)

    """
    # Populate offset array
    offset[ecount[0]] = ccount[0]
    
    # Populate cell type array and 
    if lin:
        cells[ccount[0]] = 4; ccount[0] += 1
        cell_type[ecount[0]] = vtktetnum
        
    else:
        cells[ccount[0]] = 10; ccount[0] += 1
        cell_type[ecount[0]] = vtkquadtetnum
        

    # Populate cell array while renumbering nodes

    # edge nodes
    # [0, 1, 2, 2, 3, 3, 3, 3]
    cells[ccount[0]] = numref[elem[i, 0]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 1]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 2]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 4]]; ccount[0] += 1
        
    # midside nodes
    if not lin:
        cells[ccount[0]] = numref[elem[i,  8]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i,  9]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 11]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 16]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 17]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 18]]; ccount[0] += 1

    # increment element counter
    ecount[0] += 1
        

cdef inline void StorePyr(long [::1] offset, long *ecount, long *ccount, 
                          long [::1] cells, uint8 [::1] cell_type,
                          long [::1] numref, int [:, ::1] elem, int i, int lin):
    """
    Stores pyramid element in vtk arrays.  ANSYS elements are ordered in the
    same manner as VTK.    

    VTK DOCUMENTATION
    Linear Pyramid
    The pyramid is defined by the five points (0-4) where (0,1,2,3) is the base
    of the pyramid which, using the right hand rule, forms a quadrilaterial
    whose normal points in the direction of the pyramid apex at vertex #4.
    
    Quadradic Pyramid
    The ordering of the thirteen points defining the cell is point ids
    (0-4,5-12) where point ids 0-4 are the five corner vertices of the pyramid;
    followed by eight midedge nodes (5-12). Note that these midedge nodes 
    correspond lie on the edges defined by:
    (0,1), (1,2), (2,3), (3,0), (4,5), (5,6), (6,7), (7,4), (0,4), (1,5), (2,6), (3,7)
    """

    # Populate offset and cell type arrays
    offset[ecount[0]] = ccount[0]
    
    # Populate cell array while renumbering nodes
    if lin:
        cells[ccount[0]] = 5; ccount[0] += 1
        cell_type[ecount[0]] = vtkpyrnum
    else:
        cells[ccount[0]] = 13; ccount[0] += 1
        cell_type[ecount[0]] = vtkquadpyrnum
    
    # edge nodes
    # [0, 1, 2, 3, 4, X, X, X]
    for k in range(5):
        cells[ccount[0]] = numref[elem[i, k]]
        ccount[0] += 1
        
    # Populate array
    if not lin:
        for k in range(8, 12):
            cells[ccount[0]] = numref[elem[i, k]]
            ccount[0] += 1
            
        for k in range(16, 20):
            cells[ccount[0]] = numref[elem[i, k]]
            ccount[0] += 1
    
    ecount[0] += 1

        
cdef inline void StoreWeg(long [::1] offset, long *ecount, long *ccount, 
                          long [::1] cells, uint8 [::1] cell_type,
                          long [::1] numref, int [:, ::1] elem, int i, int lin):
    """
    Stores wedge element in vtk arrays.  ANSYS elements are ordered differently
    than vtk elements.  ANSYS orders counter-clockwise and VTK orders clockwise
    
    VTK DOCUMENTATION
    Linear Wedge
    The wedge is defined by the six points (0-5) where (0,1,2) is the base of
    the wedge which, using the right hand rule, forms a triangle whose normal
    points outward (away from the triangular face (3,4,5)).

    Quadradic Wedge
    The ordering of the fifteen points defining the cell is point ids
    (0-5,6-14) where point ids 0-5 are the six corner vertices of the wedge,
    defined analogously to the six points in vtkWedge (points (0,1,2) form the
    base of the wedge which, using the right hand rule, forms a triangle whose
    normal points away from the triangular face (3,4,5)); followed by nine
    midedge nodes (6-14). Note that these midedge nodes correspond lie on the
    edges defined by :
    (0,1), (1,2), (2,0), (3,4), (4,5), (5,3), (0,3), (1,4), (2,5)
    """

    # Populate offset and cell type arrays
    offset[ecount[0]] = ccount[0]
    
    if lin:
        cells[ccount[0]] = 6; ccount[0] += 1
        cell_type[ecount[0]] = vtkwegnum

    else:
        cells[ccount[0]] = 15; ccount[0] += 1
        cell_type[ecount[0]] = vtkquadwegnum

    # Populate cell array while renumbering nodes
    # [0, 1, 2, 2, 3, 4, 5, 5]
    cells[ccount[0]] = numref[elem[i, 2]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 1]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 0]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 6]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 5]]; ccount[0] += 1
    cells[ccount[0]] = numref[elem[i, 4]]; ccount[0] += 1

    if not lin:
        # midside nodes
        cells[ccount[0]] = numref[elem[i,  9]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i,  8]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 11]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 13]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 12]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 15]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 18]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 17]]; ccount[0] += 1
        cells[ccount[0]] = numref[elem[i, 16]]; ccount[0] += 1

    ecount[0] += 1
    



cdef inline void StoreHex(long [::1] offset, long *ecount, long *ccount, 
                          long [::1] cells, uint8 [::1] cell_type,
                          long [::1] numref, int [:, ::1] elem, int i, int lin):
    """
    Stores hexahedral element in vtk arrays.  ANSYS elements are ordered in the
    same manner as VTK.    
    
    VTK DOCUMENTATION
    Linear Hexahedral
    The hexahedron is defined by the eight points (0-7) where (0,1,2,3) is the
    base of the hexahedron which, using the right hand rule, forms a
    quadrilaterial whose normal points in the direction of the opposite face
    (4,5,6,7).

    Quadradic Hexahedral
    The ordering of the twenty points defining the cell is point ids
    (0-7, 8-19) where point ids 0-7 are the eight corner vertices of the cube;
    followed by twelve midedge nodes (8-19)
    Note that these midedge nodes correspond lie on the edges defined by:
    Midside   Edge nodes
    8         (0, 1)
    9         (1, 2)
    10        (2, 3)
    11        (3, 0)
    12        (4, 5)
    13        (5, 6)
    14        (6, 7)
    15        (7, 4)
    16        (0, 4)
    17        (1, 5)
    18        (2, 6)
    19        (3, 7)

    """

    # Populate offset, cell type, and cell arrays while renumbering nodes
    offset[ecount[0]] = ccount[0]
    if lin:
        cells[ccount[0]] = 8; ccount[0] += 1
        cell_type[ecount[0]] = vtkhexnum

        for k in range(8):
            cells[ccount[0]] = numref[elem[i, k]]
            ccount[0] += 1

    else:
        cells[ccount[0]] = 20; ccount[0] += 1
        cell_type[ecount[0]] = vtkquadhexnum
        
        for k in range(20):
            cells[ccount[0]] = numref[elem[i, k]]
            ccount[0] += 1

    ecount[0] += 1
        
        
#==============================================================================
# Parse raw data
#==============================================================================
def Parse(raw, pyforce_linear):
    """
    Parses raw cdb data from downstream conversion to a vtk unstructured grid
    """
    cdef int force_linear = pyforce_linear    
    
    cdef long [:, ::1] ekey = raw['ekey']
    cdef int [:, ::1] elem = raw['elem']
    cdef int [::1] etype = raw['etype']
    cdef int [::1] nnum = raw['nnum']
    
    cdef int i, j, k, lin
    cdef int nelem = elem.shape[0]
    cdef int nnode = nnum.shape[0]

    # Find the max element type number
    cdef int nekey = ekey.shape[0]
    maxelemtype = 0
    for i in range(nekey):
        if ekey[i, 0] > maxelemtype:
            maxelemtype = ekey[i, 0]
    
    # Create an element type array for indexing
    cdef int [::1] elem_type = np.empty(maxelemtype + 1, ctypes.c_int)
    for i in range(nekey):
        elem_type[ekey[i, 0]] = ekey[i, 1]
    
    # Allocate memory for cell data
    cdef long [::1] offset = np.empty(nelem, ctypes.c_long)
    cdef uint8 [::1] cell_type = np.empty(nelem, dtype='uint8')
    
    # different array sizes depending on midside nodes
    cdef long [::1] cells = np.empty(nelem*21, ctypes.c_long)  # max cell is 20 and header is 1
    
    # Find the highest node number
    cdef int maxnodenum = 0
    for i in range(nnode):
        if nnum[i] > maxnodenum:
            maxnodenum = nnum[i]
            
    # Create reference array for node renumbering
    cdef long n
    cdef long [::1] numref = np.empty(maxnodenum + 2, ctypes.c_long)
#    numref[0]  = -1 # elements that have missing nodes written as a "0" in cdb
#
#    # elements that have missing nodes, but were written such that those
#    # missing nodes were at the end of the line and were simply not written
#    # will be stored as a -1
#    numref[-1] = -1

    # forcing all to -1 to avoid null references
    numref[:] = -1
          
    for n in range(nnode):
        numref[nnum[n]] = n
    
    # Loop through each element and check if the element type matches one this code
    # can read
    cdef long ccount = 0 # cell/offset counter
    cdef long ecount = 0 # element number counter
    for i in range(nelem):
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                # Set to read quadradic nodes
                if force_linear:
                    lin = 1
                else:
                    # Check if linear (missing midside nodes)
                    lin = elem[i, 8] == -1
                    
                ############ Read element typeA ############
                # Determine element type through logic
                if elem[i, 6] != elem[i, 7]: # check if hexahedral
                    StoreHex(offset, &ecount, &ccount, cells, cell_type, numref,
                             elem, i, lin)

                elif elem[i, 5] != elem[i, 6]: # check if wedge
                    StoreWeg(offset, &ecount, &ccount, cells, cell_type, numref,
                             elem, i, lin)
                    
                elif elem[i, 2] != elem[i, 3]: # check if pyramid
                    StorePyr(offset, &ecount, &ccount, cells, cell_type, numref,
                             elem, i, lin)
                    
                else: # if tetrahedral
                    StoreTet(offset, &ecount, &ccount, cells, cell_type, numref,
                             elem, i, lin)

                break # Continue to next element
                
        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                if force_linear:
                    lin = 1
                elif elem[i, 8] != -1:
                    lin = 0
                    
                StoreTet_TypeB(offset, &ecount, &ccount, cells, cell_type, 
                               numref, elem, i, lin)
                               
                break # Continue to next element
                    
                    
    # Return spliced arrays
    return np.asarray(cells[:ccount]), np.asarray(offset[:ecount]), \
           np.asarray(cell_type[:ecount]), np.asarray(numref)
    
    
#==============================================================================
# For FEM    
#==============================================================================
## Define cell edges ##
cdef int [6][2] tetedges
tetedges[0][0] = 0
tetedges[0][1] = 1

tetedges[1][0] = 1
tetedges[1][1] = 2

tetedges[2][0] = 0
tetedges[2][1] = 2

tetedges[3][0] = 0
tetedges[3][1] = 3

tetedges[4][0] = 1
tetedges[4][1] = 3

tetedges[5][0] = 2
tetedges[5][1] = 3


cdef int [8][2] pyredges
pyredges[0][0] = 0
pyredges[0][1] = 1

pyredges[1][0] = 1
pyredges[1][1] = 2

pyredges[2][0] = 2
pyredges[2][1] = 3

pyredges[3][0] = 0
pyredges[3][1] = 3

pyredges[4][0] = 0
pyredges[4][1] = 4

pyredges[5][0] = 1
pyredges[5][1] = 4

pyredges[6][0] = 2
pyredges[6][1] = 4

pyredges[7][0] = 3
pyredges[7][1] = 4


cdef int wegedges[9][2]
wegedges[0][0] = 0
wegedges[0][1] = 1

wegedges[1][0] = 1
wegedges[1][1] = 2

wegedges[2][0] = 2
wegedges[2][1] = 0

wegedges[3][0] = 0
wegedges[3][1] = 3

wegedges[4][0] = 1
wegedges[4][1] = 4

wegedges[5][0] = 2
wegedges[5][1] = 5

wegedges[6][0] = 3
wegedges[6][1] = 4

wegedges[7][0] = 4
wegedges[7][1] = 5

wegedges[8][0] = 3
wegedges[8][1] = 5


cdef int hexedges[12][2]
hexedges[0][0] = 0
hexedges[0][1] = 1

hexedges[1][0] = 1
hexedges[1][1] = 2

hexedges[2][0] = 2
hexedges[2][1] = 3

hexedges[3][0] = 0
hexedges[3][1] = 3

hexedges[4][0] = 0
hexedges[4][1] = 4

hexedges[5][0] = 1
hexedges[5][1] = 5

hexedges[6][0] = 2
hexedges[6][1] = 6

hexedges[7][0] = 3
hexedges[7][1] = 7

hexedges[8][0] = 4
hexedges[8][1] = 5

hexedges[9][0] = 5
hexedges[9][1] = 6

hexedges[10][0] = 6
hexedges[10][1] = 7

hexedges[11][0] = 4
hexedges[11][1] = 7

    
# ANSYS cell numbering for TypeA elements

# Wedge element
# ANSYS defines a wege as [0, 1, 2, 2, 3, 4, 5, 5]
# but VTK needs [2, 1, 0, 6, 5, 4] due to the ordering of the faces
cdef int [6] wegnum
wegnum[0] = 2
wegnum[1] = 1
wegnum[2] = 0
wegnum[3] = 6
wegnum[4] = 5
wegnum[5] = 4


# Tetrahedral element
cdef int [4] tetnum
tetnum[0] = 0
tetnum[1] = 1
tetnum[2] = 2
tetnum[3] = 4


# Midside node indices for a type A element
cdef int [12][2] typeAmidind
typeAmidind [0][0] = 0
typeAmidind [0][1] = 1

typeAmidind [1][0] = 1
typeAmidind [1][1] = 2

typeAmidind [2][0] = 2
typeAmidind [2][1] = 3

typeAmidind [3][0] = 0
typeAmidind [3][1] = 3

typeAmidind [4][0] = 4
typeAmidind [4][1] = 5

typeAmidind [5][0] = 5
typeAmidind [5][1] = 6

typeAmidind [6][0] = 6
typeAmidind [6][1] = 7

typeAmidind [7][0] = 4
typeAmidind [7][1] = 7

typeAmidind [8][0] = 0
typeAmidind [8][1] = 4

typeAmidind [9][0] = 1
typeAmidind [9][1] = 5

typeAmidind [10][0] = 2
typeAmidind [10][1] = 6

typeAmidind [11][0] = 3
typeAmidind [11][1] = 7

# typeA midside node indices are shifted by 8
cdef int [9] wegmid
wegmid[0] = 0
wegmid[1] = 1
wegmid[2] = 3
wegmid[3] = 4
wegmid[4] = 5
wegmid[5] = 7
wegmid[6] = 8
wegmid[7] = 9
wegmid[8] = 10


cdef int [8] pyrmid
pyrmid[0] = 0
pyrmid[1] = 1
pyrmid[2] = 2
pyrmid[3] = 3
pyrmid[4] = 8
pyrmid[5] = 9
pyrmid[6] = 10
pyrmid[7] = 11

cdef int [6] tetmid
tetmid[0] = 0
tetmid[1] = 1
tetmid[2] = 3
tetmid[3] = 8
tetmid[4] = 9
tetmid[5] = 10


cdef int [6][2] typeBmidind
typeBmidind[0][0] = 0
typeBmidind[0][1] = 1

typeBmidind[1][0] = 1
typeBmidind[1][1] = 2

typeBmidind[2][0] = 0
typeBmidind[2][1] = 2

typeBmidind[3][0] = 0
typeBmidind[3][1] = 3

typeBmidind[4][0] = 1
typeBmidind[4][1] = 3

typeBmidind[5][0] = 2
typeBmidind[5][1] = 3

cdef int [6] tetmidB
tetmidB[0] = 4
tetmidB[1] = 5
tetmidB[2] = 6
tetmidB[3] = 7
tetmidB[4] = 8
tetmidB[5] = 9


def ParseForFEM(raw):
    """
    Alternative approach for parsing the raw data from a cdb
    
    Colapses node numbering to 0 for downstream use and generates several other
    arrays to be used for FEA.
    """
    
    # Pull data from 
    cdef long [:, ::1] ekey = raw['ekey']
    cdef int [:, ::1] elem = raw['elem']
    cdef int [::1] etype = raw['etype']
    cdef int [::1] nnum = raw['nnum']
    
    cdef int i, j, k
    cdef int nelem = elem.shape[0]
    cdef int nnode = nnum.shape[0]

    # Find the max element type number
    cdef int nekey = ekey.shape[0]
    maxelemtype = 0
    for i in range(nekey):
        if ekey[i, 0] > maxelemtype:
            maxelemtype = ekey[i, 0]
    
    # Create an element type array for indexing
    cdef int [::1] elem_type = np.empty(maxelemtype + 1, ctypes.c_int)
    for i in range(nekey):
        elem_type[ekey[i, 0]] = ekey[i, 1]
    
    # Allocate memory for cell data
    cdef long [::1] offset = np.empty(nelem, ctypes.c_long)
    cdef uint8 [::1] cell_type = np.empty(nelem, dtype='uint8')
    
    # different array sizes depending on midside nodes
    cdef long [::1] cells = np.empty(nelem*9, ctypes.c_long)  # max cell is 8 and header is 1
    
    # Find the highest node number
    cdef int maxnodenum = 0
    for i in range(nnode):
        if nnum[i] > maxnodenum:
            maxnodenum = nnum[i]
    
    # Create reference array for node renumbering
    cdef uint8 [::1] nodeused = np.zeros(maxnodenum + 1, dtype='uint8')

    # Loop through to see which nodes will be used
    for i in range(nelem):
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                ############ Read element typeA ############
                # Determine element type through logic
                if elem[i, 6] != elem[i, 7]: # check if hexahedral
                    for k in range(8): # Indicate that these nodes are used
                        nodeused[elem[i, k]] = 1
                    
                elif elem[i, 5] != elem[i, 6]: # check if wedge
                    for k in range(6): # Indicate that these nodes are used
                        nodeused[elem[i, wegnum[k]]] = 1
                    
                elif elem[i, 2] != elem[i, 3]: # check if pyramid
                    for k in range(5): # Indicate that these nodes are used
                        nodeused[elem[i, k]] = 1
                    
                else: # if tetrahedral
                    for k in range(4): # Indicate that these nodes are used
                        nodeused[elem[i, wegnum[k]]] = 1
                
                break # Continue to next element
                
        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                for k in range(4): # Indicate that these nodes are used
                    nodeused[elem[i, k]] = 1
                break
                    
    # sum nodes used
    cdef int c = 0
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            c += 1

    # Track original node numbering        
    cdef long [::1] orignum = np.empty(nnode, ctypes.c_long)
    c = 0
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            orignum[c] = i
            c += 1
        
    # Node reference array relating original node numbering to new node
    # numbering
    cdef long [::1] numref = np.empty(maxnodenum + 1, ctypes.c_long)
    c = 0 # reset counter (this becomes the number of valid nodes)
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            numref[i] = c
            c += 1 # increment counter for next node
        else:
            numref[i] = -1

    # Create edges and square cell type array
    cdef long [:, ::1] edges = np.empty((nelem*12, 2), ctypes.c_long)
    cdef int [:, ::1] cellarr = np.zeros((nelem, 8), ctypes.c_int)
    cdef int [::1] ncellpts = np.empty(nelem, ctypes.c_int)
    
    # Track if elements are used
    cdef uint8 [::1] elemused = np.zeros(nelem, dtype='uint8')
    
    # Also, track the indices of the midside nodes based on their adjcent edge
    # nodes in the new indexing
    cdef long [:, ::1] midedgeind = np.empty((nnode, 2), ctypes.c_long) # sized for maximum number of nodes, will trim later
    
    # Also track the original midside node indices
    cdef long [::1] midind = np.empty(nnode, ctypes.c_long) # sized for maximum number of nodes, will trim later
    cdef int midnode = 0
    
    # Array to track if a midside node has been stored
    cdef int [::1] midstored = np.zeros(maxnodenum + 1, ctypes.c_int)
    
    # Loop through each element and check if the element type matches one this code
    # can read
    ccount = 0 # cell/offset counter
    ecount = 0 # element number counter
    edgenum = 0
    for i in range(nelem):
        ############ Check if element type A ############
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                # Determine element type through logic
                if elem[i, 6] != elem[i, 7]: # check if hexahedral
                    # Populate offset array
                    offset[ecount] = ccount
                
                    # Populate cell arrays with renumbered nodes
                    cells[ccount] = 8; ccount += 1
                    for k in range(8):
                        cells[ccount] = numref[elem[i, k]]
                        cellarr[ecount, k] = numref[elem[i, k]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(12):
                        edges[edgenum, 0] = cellarr[ecount, hexedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, hexedges[k][1]]
                        edgenum += 1
                        
                    # Poulate midside nodes
                    for k in range(12):
                        # if it exists and has not been stored
                        if elem[i, k + 8] > 0 and not midstored[elem[i, k + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[k][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[k][1]]]
                            midind[midnode] = elem[i, k + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, k + 8]] = 1 
                            midnode += 1
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkhexnum
                    ncellpts[ecount] = 8
                    
                    
                elif elem[i, 5] != elem[i, 6]: # check if wedge
                #    Wedge: (0-5 are edge, 6-14 are midside)
                #    [0, 1, 2, 2, 3, 4, 5, 5][6, 7, -, 8, 9, 10, -, 11, 12, 13, 14, -]
                #                            [8, 9, -,11,12, 13, -, 15, 16, 17, 18, -]
                    # Populate offset array
                    offset[ecount] = ccount  
                    
                    # Populate cell arrays with renumbered nodes
                    cells[ccount] = 6; ccount += 1
                    for k in range(6):
                        cells[ccount] = numref[elem[i, wegnum[k]]]
                        cellarr[ecount, k] = numref[elem[i, wegnum[k]]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(9):
                        edges[edgenum, 0] = cellarr[ecount, wegedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, wegedges[k][1]]
                        edgenum += 1
                        
                    # Poulate midside nodes
                    for k in range(9):
                        # if it exists
                        if elem[i, wegmid[k] + 8] > 0 and not midstored[elem[i, wegmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[wegmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[wegmid[k]][1]]]
                            midind[midnode] = elem[i, wegmid[k] + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, wegmid[k] + 8]] = 1
                            midnode += 1                        
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkwegnum
                    ncellpts[ecount] = 6

                    
                elif elem[i, 2] != elem[i, 3]: # check if pyramid
                    # Pyramid (0-4 are edge, 4-12 are midside)
                    # [0, 1, 2, 3, 4, 4, 4, 4][5, 6, 7, 8, -, -, -, -, 9, 10, 11, 12]
                    #                         [8, 9,10,11, -, -, -, -,16, 17, 18, 19]
                    
                    # Populate offset array
                    offset[ecount] = ccount  
                    
                    # Populate cell array while renumbering nodes
                    cells[ccount] = 5; ccount += 1
                    for k in range(5):
                        cells[ccount] = numref[elem[i, k]]
                        cellarr[ecount, k] = numref[elem[i, k]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(8):
                        edges[edgenum, 0] = cellarr[ecount, pyredges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, pyredges[k][1]]
                        edgenum += 1                        
                        
                    # Poulate midside nodes
                    for k in range(8):
                        # if it exists
                        if elem[i, pyrmid[k] + 8] > 0 and not midstored[elem[i, pyrmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[pyrmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[pyrmid[k]][1]]]
                            midind[midnode] = elem[i, pyrmid[k] + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, pyrmid[k] + 8]] = 1
                            midnode += 1                             
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkpyrnum
                    ncellpts[ecount] = 5
                    
                    
                else: # if tetrahedral
                    # Tetrahedral (0-3 are edge, 4-9 are midside)
                    # [0, 1, 2, 2, 3, 3, 3, 3][4, 5, -, 6, -, -, -, -, 7, 8, 9, -]
                    #                         [8, 9, -,11, -, -, -, -,16,17,18, -]
                    
                    # Populate offset array
                    offset[ecount] = ccount
                    
                    # Populate cell arrcdef long ay while renumbering nodes
                    cells[ccount] = 4; ccount += 1
                    for k in range(4):
                        cells[ccount] = numref[elem[i, tetnum[k]]]
                        cellarr[ecount, k] = numref[elem[i, tetnum[k]]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(6):
                        edges[edgenum, 0] = cellarr[ecount, tetedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, tetedges[k][1]]
                        edgenum += 1                        
                        
                    # Poulate midside nodes
                    for k in range(6):
                        # if it exists
                        if elem[i, tetmid[k] + 8] > 0 and not midstored[elem[i, tetmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[tetmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[tetmid[k]][1]]]
                            midind[midnode] = elem[i, tetmid[k] + 8]
                            
                            # track as stored and increment
                            midstored[elem[i, tetmid[k] + 8]] = 1
                            midnode += 1
                                                
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtktetnum
                    ncellpts[ecount] = 4
                    
                # Increment element counter and track that it is used
                ecount += 1   
                elemused[i] = 1
                
                break # Continue to next element
                
        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                ############ Read element typeB ############
                # Tetrahedral (with 0-3 edge and 4 - 9 midside)
                # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                # Populate offset array
                offset[ecount] = ccount  
    
                # Populate cell array while renumbering nodes
                cells[ccount] = 4; ccount += 1
                for k in range(4):
                    cells[ccount] = numref[elem[i, k]]
                    cellarr[ecount, k] = numref[elem[i, k]]
                    ccount += 1
                
                # Populate edges
                for k in range(6):
                    edges[edgenum, 0] = cellarr[ecount, tetedges[k][0]]
                    edges[edgenum, 1] = cellarr[ecount, tetedges[k][1]]
                    edgenum += 1                            
                
                # Poulate midside nodes
                for k in range(6):
                    # if it exists
                    if elem[i, tetmidB[k]] > 0 and not midstored[elem[i, tetmidB[k]]]:
                        
                        # Add its adjcent nodes to the index array tracker
                        midedgeind[midnode, 0] = numref[elem[i, typeBmidind[k][0]]]
                        midedgeind[midnode, 1] = numref[elem[i, typeBmidind[k][1]]]
                        midind[midnode] = elem[i, tetmidB[k]]

                        # track as stored and increment
                        midstored[elem[i, tetmidB[k]]] = 1
                        midnode += 1

                # Populate cell type array and number of cell points
                cell_type[ecount] = vtktetnum
                ncellpts[ecount] = 4
                
                # Increment element counter and track that it is used
                ecount += 1   
                elemused[i] = 1                    
                break # proceed to next element                     
                
                    
    # Regenerate node list
    cdef double [:, ::1] oldnodes = raw['nodes'] # c is the number of valid edge nodes
    cdef double [:, ::1] nodes = np.empty((c, 6 )) # c is the number of valid edge nodes
    cdef long [::1] nodenum = np.zeros(c, dtype=ctypes.c_long) # track if nodes are used

    cdef int n = 0 # index for old node numbering system
    for i in range(c):
        # Increment until matched
        while nnum[n] != orignum[i]:
            n += 1
        
        # Copy this node and mark as used
        for j in range(6):
            nodes[i, j] = oldnodes[n, j]
        nodenum[i] = n
                    
                    
    # Regenerate node components
    np_numref = np.asarray(numref)
    node_comps = {}
    for comp in raw['node_comps']:
        t = np_numref[raw['node_comps'][comp]]
        t = t[t != -1]
        if len(t):
            node_comps[comp] = t
    
    # Return spliced arrays
    return {'cells': np.asarray(cells[:ccount]),
            'offset': np.asarray(offset[:ecount]),
            'cell_type': np.asarray(cell_type[:ecount]),
            'edges': np.asarray(edges[:edgenum]),
            'cellarr': np.asarray(cellarr[:ecount]),
            'ncellpts': np.asarray(ncellpts[:ecount]),
            'orignode': np.asarray(orignum)[:c],
            'elemused': np.asarray(elemused),
            'nodenum': np.asarray(nodenum),
            'nodes': np.asarray(nodes),
            'node_comps': node_comps,
            'midnum': np.asarray(midind[:midnode]),
            'midside': np.asarray(midedgeind[:midnode])}
    



"""
Module to read ANSYS ASCII block formatted CDB files

USAGE

# load module
from ANSYScdb import CDB

# load ANSYS cdb file
cdb = CDB.Reader('example.cdb')

# Parse the raw data into a VTK unstructured grid
cdb.ParseVTK()

# Plot the result
cdb.Plot()

"""
import warnings
import numpy as np

# Attempt to load VTK dependent modules
try:
    import vtk
    import vtkInterface
    vtk_loaded = True
except:
    warnings.warn('Unable to load vtk dependent modules')
    vtk_loaded = False

# Cython modules
try:
    from pyansys import _reader
    from pyansys import _relaxmidside
    from pyansys import CDBparser
    cython_loaded = True
except:
    warnings.warn('Unable to load Cython modules')
    cython_loaded = False
    
from pyansys import PythonReader
from pyansys import PythonParser


class ReadArchive(object):
    """ FEM object """
    
    def __init__(self, filename='', use_cython=True, raw=None):
        """
        Initialize cdb object by reading raw cdb from file
        
        INPUTS:
        filename (string):
            filename of block formatted cdb file
            
        use_cython (bool optional):
            boolean flag to use cython reader defaults to True

        raw (dictonary optional):
            dictionary of raw data
    
        """
        
        if raw and not filename:
            # Load raw data exterinally
            self.raw = raw
            return
        
        # Defaults to cython reader if user selects it
        if use_cython and cython_loaded:
            self.raw = _reader.Read(filename)

        # Python reader for debug purposes
        else:
            self.raw = PythonReader.Read(filename)


    def ParseVTK(self, use_cython=True, force_linear=False):
        """
        Parses raw data from cdb file to VTK format.  Creates unstructured grid
        as self.uGrid
        
        Paramters
        ---------
        use_cython : bool, optional
            Select between cython parser vs. python.  Default True.
            
        force_linear : bool, optional
            This parser creates quadradic elements if available.  Set this to
            True to always create linear elements.  Defaults to False.
            
            
        Returns
        -------
        uGrid : vtk.vtkUnstructuredGrid
            VTK unstructured grid from archive file.
        
        """
        
        if not vtk_loaded:
            raise Exception('Unable to load VTK module.  Cannot parse raw cdb data')
            
           
        if self.CheckRaw():
            raise Exception('Missing key data.  Cannot parse into unstructured grid')            
           
        # Convert to vtk style arrays
        if use_cython and cython_loaded:
            cells, offset, cell_type, numref = CDBparser.Parse(self.raw,
                                                               force_linear)
            
        else:
            cells, offset, cell_type, numref = PythonParser.Parse(self.raw, 
                                                                  force_linear)

        # catch bug
        cells[cells > numref.max()] = -1
            
        # Check for missing midside nodes
        possible_merged = False
        if force_linear or np.all(cells != -1):
            nodes = self.raw['nodes'][:, :3].copy()
            nnum = self.raw['nnum']
            
        else:
            mask = cells == -1
            
            nextra = mask.sum()
            maxnum = numref.max() + 1
            cells[mask] = np.arange(maxnum, maxnum + nextra)
            
            nnodes = self.raw['nodes'].shape[0]
            nodes = np.zeros((nnodes + nextra, 3))
            nodes[:nnodes] = self.raw['nodes'][:, :3]
            
            # Add extra node numbers
            nnum = np.hstack((self.raw['nnum'], np.ones(nextra, np.int32)*-1))
            
            if cython_loaded:
                # Set new midside nodes directly between their edge nodes
                temp_nodes = nodes.copy()
                _relaxmidside.ResetMidside(cells, temp_nodes)
                nodes[nnodes:] = temp_nodes[nnodes:]
                
                possible_merged = True
                
        # Create unstructured grid
        uGrid = vtkInterface.MakeuGrid(offset, cells, cell_type, nodes)
        
        # Store original ANSYS cell and node numbering
        uGrid.AddPointScalars(nnum, 'ANSYSnodenum')
        
        # Add node components to unstructured grid
        ibool = np.empty(uGrid.GetNumberOfPoints(), dtype=np.int8)
        for comp in self.raw['node_comps']:
            ibool[:] = 0 # reset component array

            # Convert to new node numbering
            nodenum = numref[self.raw['node_comps'][comp]]
            
            ibool[nodenum] = 1
            uGrid.AddPointScalars(ibool, comp.strip())
            
        # Add tracker for original node numbering
#        npoints = uGrid.GetNumberOfPoints()
#        uGrid.AddPointScalars(np.arange(npoints), 'VTKorigID')
        
        # merge duplicate points
        if possible_merged:
            vtkappend = vtk.vtkAppendFilter()
            vtkappend.AddInputData(uGrid)
            vtkappend.MergePointsOn()
            vtkappend.Update()
            uGrid = vtkappend.GetOutput()
            vtkInterface.AddFunctions(uGrid)
        
        # Add tracker for original node numbering
        npoints = uGrid.GetNumberOfPoints()
        uGrid.AddPointScalars(np.arange(npoints), 'VTKorigID')
        
        self.vtkuGrid = uGrid
        
        return uGrid
        
        
    def ParseFEM(self, use_cython=True, raw=None):
        """ Parses raw data from cdb file to VTK format """
        if not vtk_loaded:
            raise Exception('Unable to load VTK module.  Cannot parse raw cdb data')
            return
            
        if self.CheckRaw():
            raise Exception('Missing key data.  Cannot parse into unstructured grid.')            
            
        # Convert to vtk style arrays
        if use_cython and cython_loaded:
            self.data = CDBparser.ParseForFEM(self.raw)
        else:
            self.data = PythonParser.ParseForFEM(self.raw)
            
        # Create unstructured grid
        self.uGrid = vtkInterface.MakeuGrid(self.data['offset'], self.data['cells'], 
                                            self.data['cell_type'],
                                            self.data['nodes'][:, :3])

        # Store original ANSYS cell and node numbering
        self.uGrid.AddPointScalars(self.data['orignode'], 'ANSYSnodenum')

        # Extract ANSYS element numbering and store
        ansyselem = self.raw['enum'].compress(self.data['elemused'])
        self.uGrid.AddCellScalars(ansyselem, 'ANSYSelemnum')

        # Add node components to unstructured grid
        ibool = np.empty(self.uGrid.GetNumberOfPoints(), dtype=np.int8)
        for comp in self.data['node_comps']:
            ibool[:] = 0 # reset component array
            ibool[self.data['node_comps'][comp]] = 1
            self.uGrid.AddPointScalars(ibool, comp.strip())
            
        # Add tracker for original node numbering
        npoints = self.uGrid.GetNumberOfPoints()
        self.uGrid.AddPointScalars(np.arange(npoints), 'VTKorigID')
                                  
        return self.data, self.uGrid, self.data['cellarr'], self.data['ncellpts']
        
        
    def AddThickness(self):
        """
        Adds 'thickness' point scalars to uGrid
        
        Assumes that thickness is stored as SURF154 elements
        
        """
        nnum = self.uGrid.GetPointScalars('ANSYSnodenum')        
        t = ExtractThickness(self.raw)[nnum]
        
        self.uGrid.AddPointScalars(t, 'thickness', False)
        self.hasthickness = True
        

    def Plot(self):
        """ Plot unstructured grid """
        if not vtk_loaded:
            raise Exception('VTK not loaded')

        if hasattr(self, 'vtkuGrid'):
            grid = self.vtkuGrid

        elif hasattr(self, 'uGrid'):
            grid = self.uGrid
        
        else:
            raise Exception('Unstructred grid not generated.  Run ParseVTK or ParseFEM first.')

        if not grid.GetNumberOfCells():
            raise Exception('Unstructured grid contains no cells')
        grid.Plot()


    def CheckRaw(self):
        """ Check if raw data can be converted into an unstructured grid """
        try:
            self.raw['elem'][0, 0]
            self.raw['enum'][0]
        except:
            return 1

        return 0
    
    
    def SaveAsVTK(self, filename, binary=True):
        """
        Writes the ANSYS FEM as a vtk file.
        
        The file extension will select the type of writer to use.  *.vtk will
        use the legacy writer, while *.vtu will select the VTK XML writer.
        
        Run ParseFEM or ParseVTK before running this to generate the vtk object
        
        Parameters
        ----------
        filename : str
            Filename of grid to be written.  The file extension will select the 
            type of writer to use.  *.vtk will use the legacy writer, 
            while *.vtu will select the PVTK XML writer
        binary : bool, optional
            Writes as a binary file by default.  Set to False to write ASCII
            
        Returns
        -------
        None
            
        Notes
        -----
        Binary files write much faster than ASCII, but binary files written on
        one  system may not be readable on other systems.  Binary can only be
        selected for the legacy writer.
        
        """    
        
        # Check if the unstructured grid exists
        if not hasattr(self, 'uGrid'):
            raise Exception('Run ParseFEM or ParseVTK first.')
            
        # Write the grid
        self.uGrid.WriteGrid(filename, binary)
            
            
def ExtractThickness(raw):
    """
    Extract thickness from raw element data:

    Assumes that thickness is stored as a real constant for SURF154 elements

    The "thickness" of nodes belonging to multiple SURF154 elements will is 
    averaged   
    
    """
    
    ekey = raw['ekey']
    etype = raw['etype']
    nnum = raw['nnum']
    e_rcon = raw['e_rcon'] # real constants
    rdat = raw['rdat']
    rnum = raw['rnum']
    
    # Assemble node thickness array (nodes belonging to SURF154)
    ety = np.in1d(ekey[:, 1], 154)
    maskC = np.in1d(etype, ekey[ety, 0])
    
    # Create thickness array
    maxnode = nnum.max() + 1
    t = np.zeros(maxnode)
    a = np.zeros(maxnode, np.int32) # number of entries in thickness array
    
    if np.any(maskC):

        # Reduce element matrix to maskC elements and first four nodes
        elem = raw['elem'][maskC, :8] 
        e_rcon = e_rcon[maskC]

        # Get the nodes of the elements matching each real constant
        for i in range(len(rnum)):
            # Get all surf154 elements matching the real constant
            idx = elem[e_rcon == rnum[i]] # get node indices
            
            idx = idx[idx != -1]
                      
            # Add thickness            
            t[idx] += rdat[i][6]
            a[idx] += 1
        
        # normalize thickness by number of entires
        a[a == 0] = 1 # avoid divide by zero
        t /= a
    
    return t


def Unique_Rows(a):
    """ Returns unique rows of a and indices of those rows """
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx = np.unique(b, return_index=True)
    
    return a[idx], idx


import numpy as np

# VTK cell type numbering from header file

# Linear elements                   
vtktetnum = 10 # VTK_TETRA = 10
vtkpyrnum = 14 # VTK_PYRAMID = 14,
vtkwegnum = 13 # VTK_WEDGE = 13
vtkhexnum = 12 # VTK_HEXAHEDRON = 12

# Quadradic elements
vtkquadtetnum = 24 # VTK_QUADRATIC_TETRA = 24
vtkquadpyrnum = 27 # VTK_QUADRATIC_PYRAMID = 27
vtkquadwegnum = 26 # VTK_QUADRATIC_WEDGE = 26
vtkquadhexnum = 25 # VTK_QUADRATIC_HEXAHEDRON = 25
                   
# ANSYS mixed mesh types             
typeA = np.empty(4, np.int64)
typeA[0] = 45
typeA[1] = 95
typeA[2] = 185
typeA[3] = 186

# ANSYS tetrahedral mesh
typeB = np.empty(2, np.int64)
typeB[0] = 92
typeB[1] = 187

#==============================================================================
# Cell storage
#==============================================================================
def StoreTet_TypeB(offset, ecount, ccount, cells, cell_type, numref, elem, i,
                   lin):
    """
    Stores tetrahedral element in vtk arrays.  ANSYS elements are ordered
    the same as vtk elements.
    
    VTK DOCUMENTATION
    The tetrahedron is defined by the four points (0-3); where (0,1,2) is the
    base of the tetrahedron which, using the right hand rule, forms a triangle
    whose normal points in the direction of the fourth point.
    
    For quadradic
    The cell includes a mid-edge node on each of the size edges of the 
    tetrahedron. The ordering of the ten points defining the cell is point ids
    (0-3,4-9) where ids 0-3 are the four tetra vertices; and point ids 4-9 are
    the midedge nodes between
    (0,1), (1,2), (2,0), (0,3), (1,3), and (2,3)

    """
    # Populate offset array
    offset[ecount] = ccount
    
    if lin:
        cells[ccount] = 4; ccount += 1
        
        # Populate cell array while renumbering nodes
        for j in range(4):
            cells[ccount] = numref[elem[i, j]]; ccount += 1
        
        # Populate cell type array
        cell_type[ecount] = vtktetnum
        
    else:
        cells[ccount] = 10; ccount += 1
        
        # Populate cell array while renumbering nodes
        for j in range(10):
            cells[ccount] = numref[elem[i, j]]; ccount += 1
        
        # Populate cell type array
        cell_type[ecount] = vtkquadtetnum
        
    ecount += 1
    return ccount, ecount
        

def StoreTet(offset, ecount, ccount, cells, cell_type, numref, elem, i, lin):
    """
    Stores tetrahedral element in vtk arrays.  ANSYS elements are ordered
    the same as vtk elements.
    
    VTK DOCUMENTATION:
    Linear
    The tetrahedron is defined by the four points (0-3); where (0,1,2) is the
    base of the tetrahedron which, using the right hand rule, forms a triangle
    whose normal points in the direction of the fourth point.
 
    Quadradic   
    The cell includes a mid-edge node on each of the size edges of the 
    tetrahedron. The ordering of the ten points defining the cell is point ids
    (0-3,4-9) where ids 0-3 are the four tetra vertices; and point ids 4-9 are
    the midedge nodes between
    (0,1), (1,2), (2,0), (0,3), (1,3), and (2,3)

    """
    # Populate offset array
    offset[ecount] = ccount
    
    # Populate cell type array and 
    if lin:
        cells[ccount] = 4; ccount += 1
        cell_type[ecount] = vtktetnum
        
    else:
        cells[ccount] = 10; ccount += 1
        cell_type[ecount] = vtkquadtetnum
        

    # Populate cell array while renumbering nodes

    # edge nodes
    # [0, 1, 2, 2, 3, 3, 3, 3]
    cells[ccount] = numref[elem[i, 0]]; ccount += 1
    cells[ccount] = numref[elem[i, 1]]; ccount += 1
    cells[ccount] = numref[elem[i, 2]]; ccount += 1
    cells[ccount] = numref[elem[i, 4]]; ccount += 1
        
    # midside nodes
    if not lin:
        cells[ccount] = numref[elem[i,  8]]; ccount += 1
        cells[ccount] = numref[elem[i,  9]]; ccount += 1
        cells[ccount] = numref[elem[i, 11]]; ccount += 1
        cells[ccount] = numref[elem[i, 16]]; ccount += 1
        cells[ccount] = numref[elem[i, 17]]; ccount += 1
        cells[ccount] = numref[elem[i, 18]]; ccount += 1
    
    ecount += 1
    return ccount, ecount
    

def StoreWeg(offset, ecount, ccount, cells, cell_type, numref, elem, i, lin):
    """
    Stores wedge element in vtk arrays.  ANSYS elements are ordered differently
    than vtk elements.  ANSYS orders counter-clockwise and VTK orders clockwise
    
    VTK DOCUMENTATION
    Linear Wedge
    The wedge is defined by the six points (0-5) where (0,1,2) is the base of
    the wedge which, using the right hand rule, forms a triangle whose normal
    points outward (away from the triangular face (3,4,5)).

    Quadradic Wedge
    The ordering of the fifteen points defining the cell is point ids
    (0-5,6-14) where point ids 0-5 are the six corner vertices of the wedge,
    defined analogously to the six points in vtkWedge (points (0,1,2) form the
    base of the wedge which, using the right hand rule, forms a triangle whose
    normal points away from the triangular face (3,4,5)); followed by nine
    midedge nodes (6-14). Note that these midedge nodes correspond lie on the
    edges defined by :
    (0,1), (1,2), (2,0), (3,4), (4,5), (5,3), (0,3), (1,4), (2,5)
    """

    # Populate offset and cell type arrays
    offset[ecount] = ccount
    
    if lin:
        cells[ccount] = 6; ccount += 1
        cell_type[ecount] = vtkwegnum

    else:
        cells[ccount] = 15; ccount += 1
        cell_type[ecount] = vtkquadwegnum

    # Populate cell array while renumbering nodes
    # [0, 1, 2, 2, 3, 4, 5, 5]
    cells[ccount] = numref[elem[i, 2]]; ccount += 1
    cells[ccount] = numref[elem[i, 1]]; ccount += 1
    cells[ccount] = numref[elem[i, 0]]; ccount += 1
    cells[ccount] = numref[elem[i, 6]]; ccount += 1
    cells[ccount] = numref[elem[i, 5]]; ccount += 1
    cells[ccount] = numref[elem[i, 4]]; ccount += 1

    if not lin:
        # midside nodes
        cells[ccount] = numref[elem[i,  9]]; ccount += 1
        cells[ccount] = numref[elem[i,  8]]; ccount += 1
        cells[ccount] = numref[elem[i, 11]]; ccount += 1
        cells[ccount] = numref[elem[i, 13]]; ccount += 1
        cells[ccount] = numref[elem[i, 12]]; ccount += 1
        cells[ccount] = numref[elem[i, 15]]; ccount += 1
        cells[ccount] = numref[elem[i, 18]]; ccount += 1
        cells[ccount] = numref[elem[i, 17]]; ccount += 1
        cells[ccount] = numref[elem[i, 16]]; ccount += 1

    ecount += 1
    return ccount, ecount
    

def StorePyr(offset, ecount, ccount, cells, cell_type, numref, elem, i, lin):
    """
    Stores pyramid element in vtk arrays.  ANSYS elements are ordered in the
    same manner as VTK.    

    VTK DOCUMENTATION
    Linear Pyramid
    The pyramid is defined by the five points (0-4) where (0,1,2,3) is the base
    of the pyramid which, using the right hand rule, forms a quadrilaterial
    whose normal points in the direction of the pyramid apex at vertex #4.
    
    Quadradic Pyramid
    The ordering of the thirteen points defining the cell is point ids
    (0-4,5-12) where point ids 0-4 are the five corner vertices of the pyramid;
    followed by eight midedge nodes (5-12). Note that these midedge nodes 
    correspond lie on the edges defined by:
    (0, 1), (1, 2), (2, 3), (3, 0), (0, 4), (1, 4), (2, 4), (3, 4)
    """

    # Populate offset and cell type arrays
    offset[ecount] = ccount  
    
    # Populate cell array while renumbering nodes
    if lin:
        cells[ccount] = 5; ccount += 1
        cell_type[ecount] = vtkpyrnum
    else:
        cells[ccount] = 13; ccount += 1
        cell_type[ecount] = vtkquadpyrnum
    
    # edge nodes
    # [0, 1, 2, 3, 4, X, X, X]
    for k in range(5):
        cells[ccount] = numref[elem[i, k]]
        ccount += 1
        
    # Populate array
    if not lin:
        for k in range(8, 12):
            cells[ccount] = numref[elem[i, k]]
            ccount += 1
            
        for k in range(16, 20):
            cells[ccount] = numref[elem[i, k]]
            ccount += 1
    
    ecount += 1
    return ccount, ecount


def StoreHex(offset, ecount, ccount, cells, cell_type, numref, elem, i, lin):
    """
    Stores hexahedral element in vtk arrays.  ANSYS elements are ordered in the
    same manner as VTK.    
    
    VTK DOCUMENTATION
    Linear Hexahedral
    The hexahedron is defined by the eight points (0-7) where (0,1,2,3) is the
    base of the hexahedron which, using the right hand rule, forms a
    quadrilaterial whose normal points in the direction of the opposite face
    (4,5,6,7).

    Quadradic Hexahedral
    The ordering of the twenty points defining the cell is point ids
    (0-7, 8-19) where point ids 0-7 are the eight corner vertices of the cube;
    followed by twelve midedge nodes (8-19)
    Note that these midedge nodes correspond lie on the edges defined by:
    Midside   Edge nodes
    8         (0, 1)
    9         (1, 2)
    10        (2, 3)
    11        (3, 0)
    12        (4, 5)
    13        (5, 6)
    14        (6, 7)
    15        (7, 4)
    16        (0, 4)
    17        (1, 5)
    18        (2, 6)
    19        (3, 7)

    """

    # Populate offset, cell type, and cell arrays while renumbering nodes
    offset[ecount] = ccount
    if lin:
        cells[ccount] = 8; ccount += 1
        cell_type[ecount] = vtkhexnum

        for k in range(8):
            cells[ccount] = numref[elem[i, k]]
            ccount += 1

    else:
        cells[ccount] = 20; ccount += 1
        cell_type[ecount] = vtkquadhexnum
        
        for k in range(20):
            cells[ccount] = numref[elem[i, k]]
            ccount += 1

    ecount += 1
    return ccount, ecount
    
#==============================================================================
# Parses raw data without adding fem data
#==============================================================================
def Parse(raw, force_linear):
    """ Pythonic implementation of CDB parser """
    
    # Import from raw dictionary
    ekey = raw['ekey']
    elem = raw['elem']
    etype = raw['etype']
    nnum = raw['nnum']
    
    nelem = elem.shape[0]
    nnode = nnum.shape[0]

    # Find the max element type number
    nekey = ekey.shape[0]
    maxelemtype = 0
    for i in range(nekey):
        if ekey[i, 0] > maxelemtype:
            maxelemtype = ekey[i, 0]
    
    # Create an element type array for indexing
    elem_type = np.empty(maxelemtype + 1, np.int32)
    for i in range(nekey):
        elem_type[ekey[i, 0]] = ekey[i, 1]
    
    # Allocate memory for cell data
    offset = np.empty(nelem, np.int64)
    cell_type = np.empty(nelem, dtype='uint8')
    
    # different array sizes depending on midside nodes
    cells = np.empty(nelem*(20 + 1), np.int64) # max cell is 20 and header is 1
    
    # Find the highest node number
    maxnodenum = 0
    for i in range(nnode):
        if nnum[i] > maxnodenum:
            maxnodenum = nnum[i]
    
    # Create reference array for node renumbering
    numref = np.empty(maxnodenum + 1, np.int64)
    for n in range(nnode):
        numref[nnum[n]] = n
    
    # Loop through each element and check if the element type matches one this code
    # can read
    ccount = 0 # cell/offset counter
    ecount = 0 # element number counter
    for i in range(nelem):
        
        # Test for element type A
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                ##################### Read element type A #####################
                # Determine if linear
                if force_linear:
                    lin = True
                else:
                    lin = elem[i, 8] == -1
            
                if elem[i, 6] != elem[i, 7]: # if hexahedral
                    ccount, ecount = StoreHex(offset, ecount, ccount, cells, 
                                              cell_type, numref, elem, i, lin)
                                 
                elif elem[i, 5] != elem[i, 6]: # if wedge
                    ccount, ecount = StoreWeg(offset, ecount, ccount, cells, 
                                              cell_type, numref, elem, i, lin)

                elif elem[i, 2] != elem[i, 3]: # if pyramid
                    ccount, ecount = StorePyr(offset, ecount, ccount, cells, 
                                              cell_type, numref, elem, i, lin)

                else: # if tetrahedral
                    ccount, ecount = StoreTet(offset, ecount, ccount, cells, 
                                              cell_type, numref, elem, i, lin)
                                              
                break # don't continue testing element types

        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                if force_linear:
                    lin = True
                else:
                    lin = elem[i, 4] == -1

                ccount, ecount = StoreTet_TypeB(offset, ecount, ccount, cells,
                                                cell_type, numref, elem, i, 
                                                lin)
                            
                break # don't continue testing element types
                            
        if ecount > i + 1:
            break
                
    # Return spliced arrays
    return np.asarray(cells[:ccount]), np.asarray(offset[:ecount]), np.asarray(cell_type[:ecount]), np.asarray(numref)             

                   
#==============================================================================
# For FEM    
#==============================================================================
## Define cell edges ##
tetedges = np.empty((6, 2), np.int32)
tetedges[0][0] = 0
tetedges[0][1] = 1

tetedges[1][0] = 1
tetedges[1][1] = 2

tetedges[2][0] = 0
tetedges[2][1] = 2

tetedges[3][0] = 0
tetedges[3][1] = 3

tetedges[4][0] = 1
tetedges[4][1] = 3

tetedges[5][0] = 2
tetedges[5][1] = 3


pyredges = np.empty((8, 2), np.int32)
pyredges[0][0] = 0
pyredges[0][1] = 1

pyredges[1][0] = 1
pyredges[1][1] = 2

pyredges[2][0] = 2
pyredges[2][1] = 3

pyredges[3][0] = 0
pyredges[3][1] = 3

pyredges[4][0] = 0
pyredges[4][1] = 4

pyredges[5][0] = 1
pyredges[5][1] = 4

pyredges[6][0] = 2
pyredges[6][1] = 4

pyredges[7][0] = 3
pyredges[7][1] = 4

wegedges = np.empty((9, 2), np.int32)
wegedges[0][0] = 0
wegedges[0][1] = 1

wegedges[1][0] = 1
wegedges[1][1] = 2

wegedges[2][0] = 2
wegedges[2][1] = 0

wegedges[3][0] = 0
wegedges[3][1] = 3

wegedges[4][0] = 1
wegedges[4][1] = 4

wegedges[5][0] = 2
wegedges[5][1] = 5

wegedges[6][0] = 3
wegedges[6][1] = 4

wegedges[7][0] = 4
wegedges[7][1] = 5

wegedges[8][0] = 3
wegedges[8][1] = 5

hexedges = np.empty((12, 2), np.int32)
hexedges[0][0] = 0
hexedges[0][1] = 1

hexedges[1][0] = 1
hexedges[1][1] = 2

hexedges[2][0] = 2
hexedges[2][1] = 3

hexedges[3][0] = 0
hexedges[3][1] = 3

hexedges[4][0] = 0
hexedges[4][1] = 4

hexedges[5][0] = 1
hexedges[5][1] = 5

hexedges[6][0] = 2
hexedges[6][1] = 6

hexedges[7][0] = 3
hexedges[7][1] = 7

hexedges[8][0] = 4
hexedges[8][1] = 5

hexedges[9][0] = 5
hexedges[9][1] = 6

hexedges[10][0] = 6
hexedges[10][1] = 7

hexedges[11][0] = 4
hexedges[11][1] = 7

    
# ANSYS cell numbering for TypeA elements

# Wedge element
# ANSYS defines a wege as [0, 1, 2, 2, 3, 4, 5, 5]
# but VTK needs [2, 1, 0, 6, 5, 4] due to the ordering of the faces
wegnum = np.empty(6, np.int32)
wegnum[0] = 2
wegnum[1] = 1
wegnum[2] = 0
wegnum[3] = 6
wegnum[4] = 5
wegnum[5] = 4


# Tetrahedral element
tetnum = np.empty(4, np.int32)
tetnum[0] = 0
tetnum[1] = 1
tetnum[2] = 2
tetnum[3] = 4


# Midside node indices for a type A element
typeAmidind = np.empty((12, 2), np.int32)
typeAmidind [0][0] = 0
typeAmidind [0][1] = 1

typeAmidind [1][0] = 1
typeAmidind [1][1] = 2

typeAmidind [2][0] = 2
typeAmidind [2][1] = 3

typeAmidind [3][0] = 0
typeAmidind [3][1] = 3

typeAmidind [4][0] = 4
typeAmidind [4][1] = 5

typeAmidind [5][0] = 5
typeAmidind [5][1] = 6

typeAmidind [6][0] = 6
typeAmidind [6][1] = 7

typeAmidind [7][0] = 4
typeAmidind [7][1] = 7

typeAmidind [8][0] = 0
typeAmidind [8][1] = 4

typeAmidind [9][0] = 1
typeAmidind [9][1] = 5

typeAmidind [10][0] = 2
typeAmidind [10][1] = 6

typeAmidind [11][0] = 3
typeAmidind [11][1] = 7

# typeA midside node indices are shifted by 8
wegmid = np.empty(9, np.int32)
wegmid[0] = 0
wegmid[1] = 1
wegmid[2] = 3
wegmid[3] = 4
wegmid[4] = 5
wegmid[5] = 7
wegmid[6] = 8
wegmid[7] = 9
wegmid[8] = 10


pyrmid = np.empty(8, np.int32)
pyrmid[0] = 0
pyrmid[1] = 1
pyrmid[2] = 2
pyrmid[3] = 3
pyrmid[4] = 8
pyrmid[5] = 9
pyrmid[6] = 10
pyrmid[7] = 11

tetmid = np.empty(6, np.int32)
tetmid[0] = 0
tetmid[1] = 1
tetmid[2] = 3
tetmid[3] = 8
tetmid[4] = 9
tetmid[5] = 10


typeBmidind = np.empty((6, 2), np.int32)
typeBmidind[0][0] = 0
typeBmidind[0][1] = 1

typeBmidind[1][0] = 1
typeBmidind[1][1] = 2

typeBmidind[2][0] = 0
typeBmidind[2][1] = 2

typeBmidind[3][0] = 0
typeBmidind[3][1] = 3

typeBmidind[4][0] = 1
typeBmidind[4][1] = 3

typeBmidind[5][0] = 2
typeBmidind[5][1] = 3

tetmidB = np.empty(6, np.int32)
tetmidB[0] = 4
tetmidB[1] = 5
tetmidB[2] = 6
tetmidB[3] = 7
tetmidB[4] = 8
tetmidB[5] = 9


def ParseForFEM(raw):
    """
    Alternative approach for parsing the raw data from a cdb
    
    Colapses node numbering to 0 for downstream use
    """
    
    ekey = raw['ekey']
    elem = raw['elem']
    etype = raw['etype']
    nnum = raw['nnum']
    
    nelem = elem.shape[0]
    nnode = nnum.shape[0]

    # Find the max element type number
    nekey = ekey.shape[0]
    maxelemtype = 0
    for i in range(nekey):
        if ekey[i, 0] > maxelemtype:
            maxelemtype = ekey[i, 0]
    
    # Create an element type array for indexing
    elem_type = np.empty(maxelemtype + 1, np.int32)
    for i in range(nekey):
        elem_type[ekey[i, 0]] = ekey[i, 1]
    
    # Allocate memory for cell data
    offset = np.empty(nelem, np.int64)
    cell_type = np.empty(nelem, dtype='uint8')
    
    # different array sizes depending on midside nodes
    cells = np.empty(nelem*9, np.int64)  # max cell is 8 and header is 1
    
    # Find the highest node number
    maxnodenum = 0
    for i in range(nnode):
        if nnum[i] > maxnodenum:
            maxnodenum = nnum[i]
    
    # Create reference array for node renumbering
    nodeused = np.zeros(maxnodenum + 1, dtype='uint8')

    # Loop through to see which nodes will be used
    c = 0
    for i in range(nelem):
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                ############ Read element typeA ############
                # Determine element type through logic
                if elem[i, 6] != elem[i, 7]: # check if hexahedral
                    for k in range(8): # Indicate that these nodes are used
                        nodeused[elem[i, k]] = 1
                    
                elif elem[i, 5] != elem[i, 6]: # check if wedge
                    for k in range(6): # Indicate that these nodes are used
                        nodeused[elem[i, wegnum[k]]] = 1
                    
                elif elem[i, 2] != elem[i, 3]: # check if pyramid
                    for k in range(5): # Indicate that these nodes are used
                        nodeused[elem[i, k]] = 1
                    
                else: # if tetrahedral
                    for k in range(4): # Indicate that these nodes are used
                        nodeused[elem[i, wegnum[k]]] = 1
                
                break # Continue to next element
                
        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                for k in range(4): # Indicate that these nodes are used
                    nodeused[elem[i, k]] = 1
                break
                        
    # sum nodes used
    c = 0
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            c += 1

    # Track original node numbering        
    orignum = np.empty(nnode, np.int64)
    c = 0
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            orignum[c] = i
            c += 1
        
    # Node reference array relating original node numbering to new node
    # numbering
    numref = np.empty(maxnodenum + 1, np.int64)
    c = 0 # reset counter (this becomes the number of valid nodes)
    for i in range(maxnodenum + 1):
        if nodeused[i]:
            numref[i] = c
            c += 1 # increment counter for next node
        else:
            numref[i] = -1

    # Create edges and square cell type array
    edges = np.empty((nelem*12, 2), np.int64)
    cellarr = np.zeros((nelem, 8), np.int32)
    ncellpts = np.empty(nelem, np.int32)
    
    # Track if elements are used
    elemused = np.zeros(nelem, dtype='uint8')
    
    # Also, track the indices of the midside nodes based on their adjcent edge
    # nodes in the new indexing
    midedgeind = np.empty((nnode, 2), np.int64) # sized for maximum number of nodes, will trim later
    
    # Also track the original midside node indices
    midind = np.empty(nnode, np.int64) # sized for maximum number of nodes, will trim later
    midnode = 0
    
    # Array to track if a midside node has been stored
    midstored = np.zeros(maxnodenum + 1, np.int32)
    
    # Loop through each element and check if the element type matches one this code
    # can read
    ccount = 0 # cell/offset counter
    ecount = 0 # element number counter
    edgenum = 0
    for i in range(nelem):
        ############ Read element typeA ############
        for j in range(4):
            if elem_type[etype[i]] == typeA[j]:
                # Determine element type through logic
                if elem[i, 6] != elem[i, 7]: # check if hexahedral
                    # Populate offset array
                    offset[ecount] = ccount
                
                    # Populate cell arrays with renumbered nodes
                    cells[ccount] = 8; ccount += 1
                    for k in range(8):
                        cells[ccount] = numref[elem[i, k]]
                        cellarr[ecount, k] = numref[elem[i, k]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(12):
                        edges[edgenum, 0] = cellarr[ecount, hexedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, hexedges[k][1]]
                        edgenum += 1
                        
                    # Poulate midside nodes
                    for k in range(12):
                        # if it exists and has not been stored
                        if elem[i, k + 8] > 0 and not midstored[elem[i, k + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[k][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[k][1]]]
                            midind[midnode] = elem[i, k + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, k + 8]] = 1 
                            midnode += 1
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkhexnum
                    ncellpts[ecount] = 8
                    
                    
                elif elem[i, 5] != elem[i, 6]: # check if wedge
                #    Wedge: (0-5 are edge, 6-14 are midside)
                #    [0, 1, 2, 2, 3, 4, 5, 5][6, 7, -, 8, 9, 10, -, 11, 12, 13, 14, -]
                #                            [8, 9, -,11,12, 13, -, 15, 16, 17, 18, -]
                    # Populate offset array
                    offset[ecount] = ccount  
                    
                    # Populate cell arrays with renumbered nodes
                    cells[ccount] = 6; ccount += 1
                    for k in range(6):
                        cells[ccount] = numref[elem[i, wegnum[k]]]
                        cellarr[ecount, k] = numref[elem[i, wegnum[k]]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(9):
                        edges[edgenum, 0] = cellarr[ecount, wegedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, wegedges[k][1]]
                        edgenum += 1
                        
                    # Poulate midside nodes
                    for k in range(9):
                        # if it exists
                        if elem[i, wegmid[k] + 8] > 0 and not midstored[elem[i, wegmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[wegmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[wegmid[k]][1]]]
                            midind[midnode] = elem[i, wegmid[k] + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, wegmid[k] + 8]] = 1
                            midnode += 1                        
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkwegnum
                    ncellpts[ecount] = 6

                    
                elif elem[i, 2] != elem[i, 3]: # check if pyramid
                    # Pyramid (0-4 are edge, 4-12 are midside)
                    # [0, 1, 2, 3, 4, 4, 4, 4][5, 6, 7, 8, -, -, -, -, 9, 10, 11, 12]
                    #                         [8, 9,10,11, -, -, -, -,16, 17, 18, 19]
                    
                    # Populate offset array
                    offset[ecount] = ccount  
                    
                    # Populate cell array while renumbering nodes
                    cells[ccount] = 5; ccount += 1
                    for k in range(5):
                        cells[ccount] = numref[elem[i, k]]
                        cellarr[ecount, k] = numref[elem[i, k]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(8):
                        edges[edgenum, 0] = cellarr[ecount, pyredges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, pyredges[k][1]]
                        edgenum += 1                        
                        
                    # Poulate midside nodes
                    for k in range(8):
                        # if it exists
                        if elem[i, pyrmid[k] + 8] > 0 and not midstored[elem[i, pyrmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[pyrmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[pyrmid[k]][1]]]
                            midind[midnode] = elem[i, pyrmid[k] + 8]
                            
                            # Track as stored and increment
                            midstored[elem[i, pyrmid[k] + 8]] = 1
                            midnode += 1                             
                        
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtkpyrnum
                    ncellpts[ecount] = 5
                    
                    
                else: # if tetrahedral
                    # Tetrahedral (0-3 are edge, 4-9 are midside)
                    # [0, 1, 2, 2, 3, 3, 3, 3][4, 5, -, 6, -, -, -, -, 7, 8, 9, -]
                    #                         [8, 9, -,11, -, -, -, -,16,17,18, -]
                    
                    # Populate offset array
                    offset[ecount] = ccount
                    
                    # Populate cell arrcdef long ay while renumbering nodes
                    cells[ccount] = 4; ccount += 1
                    for k in range(4):
                        cells[ccount] = numref[elem[i, tetnum[k]]]
                        cellarr[ecount, k] = numref[elem[i, tetnum[k]]]
                        ccount += 1
                        
                    # Populate edges
                    for k in range(6):
                        edges[edgenum, 0] = cellarr[ecount, tetedges[k][0]]
                        edges[edgenum, 1] = cellarr[ecount, tetedges[k][1]]
                        edgenum += 1                        
                        
                    # Poulate midside nodes
                    for k in range(6):
                        # if it exists
                        if elem[i, tetmid[k] + 8] > 0 and not midstored[elem[i, tetmid[k] + 8]]:
                            # Add its adjcent nodes to the index array tracker
                            midedgeind[midnode, 0] = numref[elem[i, typeAmidind[tetmid[k]][0]]]
                            midedgeind[midnode, 1] = numref[elem[i, typeAmidind[tetmid[k]][1]]]
                            midind[midnode] = elem[i, tetmid[k] + 8]
                            
                            # track as stored and increment
                            midstored[elem[i, tetmid[k] + 8]] = 1
                            midnode += 1
                                                
                    # Populate cell type array and number of cell points
                    cell_type[ecount] = vtktetnum
                    ncellpts[ecount] = 4
                    
                # Increment element counter and track that it is used
                ecount += 1   
                elemused[i] = 1
                
                break # Continue to next element
                
        # Test for element type B
        for j in range(2):
            if elem_type[etype[i]] == typeB[j]:
                ############ Read element typeB ############
                # Tetrahedral (with 0-3 edge and 4 - 9 midside)
                # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                # Populate offset array
                offset[ecount] = ccount  
    
                # Populate cell array while renumbering nodes
                cells[ccount] = 4; ccount += 1
                for k in range(4):
                    cells[ccount] = numref[elem[i, k]]
                    cellarr[ecount, k] = numref[elem[i, k]]
                    ccount += 1
                
                # Populate edges
                for k in range(6):
                    edges[edgenum, 0] = cellarr[ecount, tetedges[k][0]]
                    edges[edgenum, 1] = cellarr[ecount, tetedges[k][1]]
                    edgenum += 1                            
                
                # Poulate midside nodes
                for k in range(6):
                    # if it exists (isn't zero because ANSYS is weird) and hasn't been
                    if elem[i, tetmidB[k]] > 0 and midstored[elem[i, tetmidB[k]]]:
                        # Add its adjcent nodes to the index array tracker
                        midedgeind[midnode, 0] = numref[elem[i, typeBmidind[k][0]]]
                        midedgeind[midnode, 1] = numref[elem[i, typeBmidind[k][1]]]
                        midind[midnode] = elem[i, tetmidB[k]]

                        # track as stored and increment
                        midstored[elem[i, tetmidB[k]]] = 1
                        midnode += 1

                
                # Populate cell type array
                cell_type[ecount] = vtkhexnum
                
                # Increment element counter and track that it is used
                ecount += 1   
                elemused[i] = 1                    
                break # proceed to next element        
                    
                    
    # Regenerate node list
    oldnodes = raw['nodes'] # c is the number of valid edge nodes
    nodes = np.empty((c, 6)) # c is the number of valid edge nodes
    nodenum = np.zeros(c, dtype=np.int64) # track if nodes are used

    n = 0 # index for old node numbering system
    for i in range(c):
        # Increment until matched
        while nnum[n] != orignum[i]:
            n += 1
        
        # Copy this node and mark as used
        for j in range(6):
            nodes[i, j] = oldnodes[n, j]
        nodenum[i] = n
                    
                    
    # Regenerate node components
    np_numref = np.asarray(numref)
    node_comps = {}
    for comp in raw['node_comps']:
        t = np_numref[raw['node_comps'][comp]]
        t = t[t != -1]
        if len(t):
            node_comps[comp] = t
    
                    
    # Return spliced arrays
    return {'cells': np.asarray(cells[:ccount]),
            'offset': np.asarray(offset[:ecount]),
            'cell_type': np.asarray(cell_type[:ecount]),
            'edges': np.asarray(edges[:edgenum]),
            'cellarr': np.asarray(cellarr[:ecount]),
            'ncellpts': np.asarray(ncellpts[:ecount]),
            'orignode': np.asarray(orignum)[:c],
            'elemused': np.asarray(elemused),
            'nodenum': np.asarray(nodenum),
            'nodes': np.asarray(nodes),
            'node_comps': node_comps,
            'midnum': np.asarray(midind[:midnode]),
            'midside': np.asarray(midedgeind[:midnode])}
    



# -*- coding: utf-8 -*-
"""
Python implementation of CDB reader
"""
import gc
import numpy as np
import copy

def Read(filename):
    """
    Reads ASCII CDB from ANSYS and returns elem_type, nnum, nodes, enum, elem,
    fixed
    
    USAGE:
    raw = ReadCDB(filename)
    
    where raw contains the fields:
    'ekey':  Element type keys
    'nnum':  Node numbering according to ANSYS
    'nodes': Node locations
    'enum':  Element numbers according to ANSYS
    'elem':  ANSYS Element matrix
    'etype': Element types
    
    """

    # Node components    
    node_comps = {}
    nfpos = -1 # NBLOCK numbering starting position
    efpos = -1 # EBLOCK numbering starting position
    
    # Read file as an iterator
    with open(filename) as fileObj:
        fpos = 0 # track position in file

        # Parse element types and read RLBLOCK
        elem_type = []
        rnum = []
        rdat = []
        for line in fileObj:
            fpos += len(line)
            if 'ET,' in line[:3]:
                elem_type.append(ParseEType(line))
                
            elif 'RLBLOCK' in line:
                rnum, rdat, fpos = RLBLOCK_Reader(fileObj, fpos, line)

            elif 'NBLOCK' in line:
                nfpos, fpos, n_isz, nfields, n_fsz, nnode = NBLOCK_Info(fileObj, line, fpos)
        
            elif 'EBLOCK' in line:
                # Identify element block location
                fpos, e_isz = EBLOCK_Info(fileObj, line, fpos)
                efpos = copy.copy(fpos) # Copy location of element block

            elif 'CMBLOCK' in line and 'NODE' in line:
                # Read node components
                comname, c_nodes = CMBLOCK_Reader(fileObj, line)
                node_comps[comname] = c_nodes 
    
    ############### End reading file as an iterator ############

    # Open file as a buffer
    fileObj = open(filename)

    # Read the block that comes first
#    if nfpos < efpos and nfpos != -1:
    if nfpos != -1:
        nnum, nodes = NBLOCK_Reader(filename, nfpos, n_isz, nfields, n_fsz, nnode)
        
    # Read element block (if it exists)
    if efpos != -1:
        etype, enum, elem, e_rcon = EBLOCK_Reader(filename, efpos, e_isz)
    else:
        etype = []
        enum = []
        elem = [[]]
        e_rcon = []
    
                                  
    # Return numpy arrays
    raw = {'rnum': np.asarray(rnum),
           'rdat': np.asarray(rdat),
           'ekey': np.array(elem_type),
           'nnum': nnum,
           'nodes': nodes,
           'enum': np.array(enum),
           'elem': np.array(elem, np.int32),
           'etype': np.array(etype, np.int32),
           'e_rcon': np.asarray(e_rcon),
           'node_comps': node_comps}

    # Garbage collect
    del elem_type, nnum, nodes, enum, elem, etype, node_comps
    gc.collect()
    
    return raw
    
    
def EBLOCK_Info(fileObj, line, fpos):
    """
    Get interger block size and store start position (fpos) for the
    element block
    """

    # Get total block size
    try:
        bsize = int(line[int(line.rfind(',')) + 1:])
        
        # Get block size and starting position of EBLOCK
        for line in fileObj:
            fpos += len(line)
            e_isz = int(line[line.find('i') + 1:line.find(')')])
            break
        
        # Seek to end of block
        c = 0
        for line in fileObj:
            c += 1
            if c > bsize:
                break
        
    except:
        # Get block size and starting position of EBLOCK
        for line in fileObj:
            fpos += len(line)
            e_isz = int(line[line.find('i') + 1:line.find(')')])
            break
        
        # Seek to end of block
        for line in fileObj:
            try:
                if int(line) == -1:
                    break
            except:
                continue
        
    return fpos, e_isz
    
    
def NBLOCK_Info(fileObj, line, fpos):
    """
    Get interger block size and store start position (fpos) for the node block
    """

    # Get total block size
    # Try to get total number of nodes
    try:
        nnode = int(line[int(line.rfind(',')) + 1:])
    except:
        nnode = -1
    
    # Get block size and starting position of EBLOCK
    for line in fileObj:
        fpos += len(line)
        i_sz = int(line[line.find('i') + 1:line.find(',')])
        nfields = int(line[line.find(',') + 1:line.find('e')])
        f_sz = int(line[line.find('e') + 1:line.find('.')])
        break
    
    # Store starting nblock position for future seek
    nstart = copy.copy(fpos)
    
    # If number of nodes has been obtained
    if nnode != -1:
        # Seek to end of block
        c = 0
        for line in fileObj:
            fpos += len(line)
            c += 1
            if c > nnode:
                break
    else:
        # Manually to end of block
        c = 0
        for line in fileObj:
            fpos += len(line)
            try:
                if int(line) == -1:
                    break
                c += 1
            except:
                continue
            
        # Number of lines is equal to the number of nodes
        nnode = c
        
    return nstart, fpos, i_sz, nfields, f_sz, nnode
    
    
def ParseEType(line):
    """ Returns element type and element number """
    elnum = int(line[3:line.find(',', 5)]) # element number
    etype = int(line[line.find(',', 5) + 1:]) # element type
    return [elnum, etype]


def RLBLOCK_Reader(fileObj, fpos, line):
    """
    Reads RLBLOCK from ANSYS *.cdb 
    
    RLBLOCK Format:
    RLBLOCK,NUMSETS,MAXSET,MAXITEMS,NPERLINE
    Format1
    Format2 
    
    Format1
    Data descriptor defining the format of the first line. For the RLBLOCK
    command, this is always (2i8,6g16.9).  The first i8 is the set number, the
    second i8 is the number of values in this set, followed by up to 6 real
    constant values.    
 
   """

    # Get number of sets
    ist = line.find(',') + 1
    ien = line[ist:].find(',') + ist
    nset = int(line[ist:ien])

    # Skip Format1 (always 2i8,6g16.9)
    for line in fileObj:
        fpos += len(line)
        break
    
    # Skip Format 2 (always 7g16.9)
    for line in fileObj:
        fpos += len(line)
        break
    
    # Read data
    rnum = []
    rdat = []
    c_set = 0
    for line in fileObj:
        fpos += len(line)
        rcon = [] # real constants
        
        c_set += 1
        if c_set > nset:
            break
        
        # Get real constant number
        rnum.append(int(line[:8]))
        
        # Number of constants
        ncon = int(line[8:16])
        
        # Get constant data
        if ncon > 6: # if multiple lines
            for i in range(6):
                rcon.append(float(line[16 + 16*i:32 + 16*i]))
                ncon -= 1
                
            # advance line
            for line in fileObj:
                fpos += len(line)
                break   
             
            # read next line
            while True:
                if ncon > 7:
                    for i in range(7):
                        rcon.append(float(line[16*i:16*(i + 1)]))
                        ncon -= 1
                        
                    # advance
                    for line in fileObj:
                        fpos += len(line)
                        break
                        
                else:
                    for i in range(ncon):
                        try:
                            rcon.append(float(line[16*i:16 + 16*i]))  
                        except:
                            rcon.append(0.0)
                        
                    break
                
        # If only one in constant data
        else:
            for i in range(ncon):
                rcon.append(float(line[16 + 16*i:32 + 16*i]))   
                ncon -= 1

        rdat.append(rcon)
                        
    return rnum, rdat, fpos

    
def CMBLOCK_Reader(fileObj, line):
    """ Reads a componenet block """
    # Get component name and nodes
    ind1 = line.find(',') + 1
    ind2 = line.find(',', ind1)
    comname = line[ind1:ind2]
    nodes = ComponentInterperter(GetComponent(fileObj))

    return comname, nodes
    
    

def NBLOCK_Reader(filename, nfpos, n_isz, nfields, n_fsz, nnode):
    """ Read NBLOCK """
    
    # Seek the start of the NBLOCK
    fileObj = open(filename)
    fileObj.seek(nfpos)
    
    nnum = np.empty(nnode, np.int32)
    nodes = np.zeros((nnode, nfields))
    for i in range(nnode):
            
        # Field 1 - Node number
        string = fileObj.read(n_isz)
        nnum[i] = int(string)
        
        # Ignore fields 2 and 3
        fileObj.read(n_isz*2) # ignored
        
        # Read fields 4-9 (or 4-6)
        for j in range(nfields):
            # Check if next character is an eol character
            bcheck = fileObj.read(1)
            if '\r' in bcheck: # if dos EOL
                # read the next eol character and continue
                fileObj.read(1)
                break
            
            elif '\n' in bcheck:
                break
            
            else:
                # Read remaining characters and convert to float
                nodes[i, j] = float(bcheck + fileObj.read(n_fsz - 1))
    
        # If all fields have been read
        if j == nfields - 1:
            bcheck = fileObj.read(1)
            if '\r' in bcheck:
                # read the next eol character and continue
                fileObj.read(1)

    # Close file                
    fileObj.close()
                
    # Return node numbers and node positions
    return nnum, nodes


def EBLOCK_Reader(filename, fpos, isz):
    """ Reads an element block
    
    Field 1 - The material number.
    Field 2 - The element type number.
    Field 3 - The real constant number.
    Field 4 - The section ID attribute (beam section) number. See elements BEAM188 and BEAM189 for more information.
    Field 5 - The element coordinate system number.
    Field 6 - The birth/death flag.
    Field 7- The solid model reference number.
    Field 8 - The element shape flag.
    Field 9 - 0
    Field 10 - The exclude key (p-elements).
    Field 11 - The element number.
    Field 12-19 - The node numbers. The next line will have the additional node numbers if there are more than eight
    """
    
    # Open and seek the start of the EBLOCK
    fileObj = open(filename)
    fileObj.seek(fpos)
    
    etype = []
    enum = []
    elem = []
    e_rcon = [] # real constant
    
    # Read on a block basis
    while True: # Run until reaching eblock
    
        # Field 1 - The material number.
        block = fileObj.read(isz) # do not store
    
        # Check if block contains an end of a line
        if '\r\n' in block:
            block = block[2:] + fileObj.read(2)
        elif '\n' in block:
            block = block[1:] + fileObj.read(1)
    
        # Check if at end of EBLOCK
        if '-1' in block:
            break
        
        # Field 2 - The element type number.
        etype.append(int(fileObj.read(isz)))
        
        # Field 3 - The real constant number.
        e_rcon.append(int(fileObj.read(isz)))

        # Skip fields 4 - 8
        fileObj.seek(isz*5, 1)
    
        # Field 9 - Number of nodes in the element
        nnode = int(fileObj.read(isz))
            
        # Field 10 - The exclude key (p-elements).
        fileObj.seek(isz, 1) # do not store
        
        # Field 11 - The element number.    
        enum.append(int(fileObj.read(isz)))
        
        # Read nodes
        eline = []
        while len(eline) < nnode:
            
            # Read a block
            block = fileObj.read(isz)
            
            # Check if a block contains an end of a line
            if '\r\n' in block:
                block = block[2:] + fileObj.read(2)
            elif '\n' in block:
                block = block[1:] + fileObj.read(1)
                
            nodenum = int(block)
            eline.append(nodenum)

    
        eline.extend((-1,)*(20 - nnode))
        elem.append(eline)
    
    fileObj.close()
        
    return etype, enum, elem, e_rcon#, nnode
    
    
def GetComponent(fileObj):
    """ Extract fixed nodes from fileObj """
    component = []
    for line in fileObj:
         
        # Get interger size
        isz = int(line[line.find('i') + 1:line.find(')')])
        
        # Number of intergers
        nblock = int(line[line.find('(') + 1:line.find('i')])
        break
            
    for line in fileObj:
        for i in range(nblock):
            try:
                component.append(int(line[isz*i:isz*(i + 1)]))
            except:
                return component # exit when complete
        
        
def GetBlockFormat(formatstring):
    """ Get node block format """
    
    # Digit Size
    d_size = int(formatstring[formatstring.find('i') + 1:formatstring.find(',')])
    f_size = int(formatstring[formatstring.find('e') + 1:formatstring.find('.')])
    
    # Node end
    idx = [d_size, d_size*3]
    
    # Node location indicies
    for i in range(3):
        idx.append(idx[-1] + f_size)
    
    return idx


def ComponentInterperter(fixed):
    """
    If a node is negative, it is describing a list from the previous node.  This is ANSYS's way of 
    saving file size when writing components.
    
    """
    
    f_new = []
    for i in range(len(fixed)):
        if fixed[i] > 0: # Append if positive
            f_new.append(fixed[i])
        else: # otherwise, append list
            f_new.append(range(abs(fixed[i - 1]) + 1, abs(fixed[i]) + 1))
    
    return np.hstack(f_new)


"""
Installation file for pyansys
"""
import os
import sys
from io import open as io_open

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

# Create a build class that includes numpy directory
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


# Check compiler and assign compile arguments accordingly
def compilerName():
  import re
  import distutils.ccompiler
  comp = distutils.ccompiler.get_default_compiler()
  getnext = False

  for a in sys.argv[2:]:
    if getnext:
      comp = a
      getnext = False
      continue
    #separated by space
    if a == '--compiler'  or  re.search('^-[a-z]*c$', a):
      getnext = True
      continue
    #without space
    m = re.search('^--compiler=(.+)', a)
    if m == None:
      m = re.search('^-[a-z]*c(.+)', a)
    if m:
      comp = m.group(1)

  return comp

# Assign arguments based on compiler
compiler = compilerName()
if compiler == 'unix' or compiler == 'msvc':
    cmp_arg = ['-O3']
else:
    cmp_arg = ['/Ox']


# Get version from version info
__version__ = None
version_file = os.path.join(os.path.dirname(__file__), 'pyansys', '_version.py')
with io_open(version_file, mode='r') as fd:
    # execute file from raw string
    exec(fd.read())
    

# Actual setup
setup(
    name='pyansys',
    packages = ['pyansys', 'pyansys.examples'],

    # Version
    version=__version__,

    description='Pythonic interface to ANSYS binary files',
    long_description=open('README.rst').read(),

    # Author details
    author='Alex Kaszynski',
    author_email='akascap@gmail.com',

    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',

        # Target audience
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',

        # MIT License
        'License :: OSI Approved :: MIT License',

        # Will work for other python 3 versions
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # Website
    url = 'https://github.com/akaszynski/pyansys',

    # Build cython modules
    cmdclass={'build_ext': build_ext},
    ext_modules=[Extension("pyansys._parsefull", 
                           ['pyansys/cython/_parsefull.pyx',
                            'pyansys/cython/parsefull.c'],
                           language='c'),
            
                 Extension("pyansys.CDBparser", 
                           ["pyansys/cython/CDBparser.pyx"],
                           language='c'),

                 Extension('pyansys._reader', 
                           ['pyansys/cython/_reader.pyx',
                            'pyansys/cython/reader.c'],
                           extra_compile_args=cmp_arg,
                           language='c',),

                 Extension("pyansys._relaxmidside", 
                           ["pyansys/cython/_relaxmidside.pyx"],
                           language='c'),
                           
                 # cell quality module
                 Extension("pyansys._cellqual", 
                           ["pyansys/cython/_cellqual.pyx"],
                           language='c'),
                           
                 Extension("pyansys._rstHelper", 
                           ["pyansys/cython/_rstHelper.pyx"],
#                           extra_compile_args=cmp_arg,
                           language='c'),
                           
                ],
                           
    keywords='vtk ANSYS cdb full rst',
                           
    package_data={'pyansys.examples': ['TetBeam.cdb', 'HexBeam.cdb', 
                                       'file.rst', 'file.full']},

    # Might work with earlier versions
    install_requires=['numpy>1.9.3', 'cython>0.23.1', 'vtkInterface>=0.1.5']

)
